# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test data for serde module."""

import inspect
import types
from typing import Any, Sequence
from unittest import mock

from meridian import backend
from meridian import constants as c
from meridian.model import prior_distribution
from meridian.model import spec
from mmm.v1.marketing import marketing_data_pb2 as marketing_pb
from mmm.v1.model.meridian import meridian_model_pb2 as meridian_pb
import numpy as np
import xarray as xr

from google.protobuf import text_format
from tensorflow.core.framework import tensor_pb2  # pylint: disable=g-direct-tensorflow-import
from tensorflow.core.framework import tensor_shape_pb2  # pylint: disable=g-direct-tensorflow-import
from tensorflow.core.framework import types_pb2  # pylint: disable=g-direct-tensorflow-import

_MediaEffectsDist = meridian_pb.MediaEffectsDistribution
_PaidMediaPriorType = meridian_pb.PaidMediaPriorType
_NonPaidTreatmentsPriorType = meridian_pb.NonPaidTreatmentsPriorType
_NonMediaBaselineFunction = (
    meridian_pb.NonMediaBaselineValue.NonMediaBaselineFunction
)

# Shared constants
_TIME_STRS = ['2021-02-01', '2021-02-08']
_MEDIA_TIME_STRS = ['2021-01-25', '2021-02-01', '2021-02-08']
_GEO_IDS = ['geo_0', 'geo_1']
_MEDIA_CHANNEL_PAID = ['ch_paid_0', 'ch_paid_1']
_MEDIA_CHANNEL_ORGANIC = ['ch_organic_0', 'ch_organic_1']
_RF_CHANNEL_PAID = ['rf_ch_paid_0', 'rf_ch_paid_1']
_RF_CHANNEL_ORGANIC = ['rf_ch_organic_0', 'rf_ch_organic_1']
_CONTROL_VARIABLES = ['control_0', 'control_1']
_NON_MEDIA_TREATMENT_VARIABLES = [
    'non_media_treatment_0',
    'non_media_treatment_1',
]


def make_tensor_shape_proto(
    dims: Sequence[int],
) -> tensor_shape_pb2.TensorShapeProto:
  tensor_shape = tensor_shape_pb2.TensorShapeProto()
  for dim in dims:
    tensor_shape.dim.append(tensor_shape_pb2.TensorShapeProto.Dim(size=dim))
  return tensor_shape


def make_tensor_proto(
    dims: Sequence[int],
    dtype: types_pb2.DataType = types_pb2.DT_FLOAT,
    bool_vals: Sequence[bool] = (),
    string_vals: Sequence[str] = (),
    tensor_content: bytes = b'',
) -> tensor_pb2.TensorProto:
  return tensor_pb2.TensorProto(
      dtype=dtype,
      tensor_shape=make_tensor_shape_proto(dims),
      bool_val=bool_vals,
      string_val=[x.encode() for x in string_vals],
      tensor_content=tensor_content,
  )


def make_sample_dataset(
    n_chains: int,
    n_draws: int,
    n_geos: int = 5,
    n_controls: int = 2,
    n_knots: int = 0,
    n_times: int = 0,
    n_media_channels: int = 0,
    n_rf_channels: int = 0,
    n_organic_media_channels: int = 0,
    n_organic_rf_channels: int = 0,
    n_non_media_channels: int = 0,
) -> xr.Dataset:
  """Creates a sample dataset with all relevant Meridian dimensions.

  Args:
    n_chains: The number of chains.
    n_draws: The number of draws per chain.
    n_geos: The number of geos.
    n_controls: The number of control variables.
    n_knots: The number of knots.
    n_times: The number of time periods.
    n_media_channels: The number of media channels.
    n_rf_channels: The number of reach and frequency channels.
    n_organic_media_channels: The number of organic media channels.
    n_organic_rf_channels: The number of organic reach and frequency channels.
    n_non_media_channels: The number of non-media channels.

  Returns:
    An xarray Dataset with sample data.
  """
  data_vars = {
      c.STEP_SIZE: (
          [c.CHAIN, c.DRAW],
          np.random.normal(size=(n_chains, n_draws)),
      ),
      c.TUNE: (
          [c.CHAIN, c.DRAW],
          np.full((n_chains, n_draws), False),
      ),
      c.TARGET_LOG_PROBABILITY_TF: (
          [c.CHAIN, c.DRAW],
          np.random.normal(size=(n_chains, n_draws)),
      ),
      c.DIVERGING: (
          [c.CHAIN, c.DRAW],
          np.full((n_chains, n_draws), False),
      ),
      c.ACCEPT_RATIO: (
          [c.CHAIN, c.DRAW],
          np.random.normal(size=(n_chains, n_draws)),
      ),
      c.N_STEPS: (
          [c.CHAIN, c.DRAW],
          np.random.normal(size=(n_chains, n_draws)),
      ),
      'is_accepted': (
          [c.CHAIN, c.DRAW],
          np.full((n_chains, n_draws), True),
      ),
  }
  coords = {
      c.CHAIN: ([c.CHAIN], np.arange(n_chains)),
      c.DRAW: ([c.DRAW], np.arange(n_draws)),
      c.GEO: ([c.GEO], np.arange(n_geos)),
      c.CONTROL_VARIABLE: (
          [c.CONTROL_VARIABLE],
          np.arange(n_controls),
      ),
  }

  if n_knots > 0:
    coords[c.KNOTS] = ([c.KNOTS], np.arange(n_knots))

  if n_times > 0:
    coords[c.TIME] = ([c.TIME], np.arange(n_times))

  if n_media_channels > 0:
    coords[c.MEDIA_CHANNEL] = (
        [c.MEDIA_CHANNEL],
        np.arange(n_media_channels),
    )

  if n_rf_channels > 0:
    coords[c.RF_CHANNEL] = (
        [c.RF_CHANNEL],
        np.arange(n_rf_channels),
    )

  if n_organic_media_channels > 0:
    coords[c.ORGANIC_MEDIA_CHANNEL] = (
        [c.ORGANIC_MEDIA_CHANNEL],
        np.arange(n_organic_media_channels),
    )

  if n_organic_rf_channels > 0:
    coords[c.ORGANIC_RF_CHANNEL] = (
        [c.ORGANIC_RF_CHANNEL],
        np.arange(n_organic_rf_channels),
    )

  if n_non_media_channels > 0:
    coords[c.NON_MEDIA_CHANNEL] = (
        [c.NON_MEDIA_CHANNEL],
        np.arange(n_non_media_channels),
    )

  return xr.Dataset(data_vars, coords=coords)


# Marketing data test data
MOCK_INPUT_DATA_NATIONAL_MEDIA_RF_NON_REVENUE = mock.MagicMock(
    kpi_type=c.NON_REVENUE,
    geo=xr.DataArray(np.array(['national_geo'])),
    time=xr.DataArray(np.array(_TIME_STRS)),
    media_time=xr.DataArray(np.array(_TIME_STRS)),
    population=xr.DataArray(
        coords={c.GEO: ['national_geo']},
        data=np.array([1.0]),
        name=c.POPULATION,
    ),
    media=xr.DataArray(
        coords={
            c.GEO: ['national_geo'],
            c.MEDIA_TIME: _TIME_STRS,
            c.MEDIA_CHANNEL: _MEDIA_CHANNEL_PAID,
        },
        data=np.array([[[41, 42], [43, 44]]]),
        name=c.MEDIA,
    ),
    media_spend=xr.DataArray(
        coords={
            c.GEO: ['national_geo'],
            c.TIME: _TIME_STRS,
            c.MEDIA_CHANNEL: _MEDIA_CHANNEL_PAID,
        },
        data=np.array([[[141, 142], [143, 144]]]),
        name=c.MEDIA_SPEND,
    ),
    media_spend_has_geo_dimension=True,
    media_spend_has_time_dimension=True,
    reach=xr.DataArray(
        coords={
            c.GEO: ['national_geo'],
            c.MEDIA_TIME: _TIME_STRS,
            c.RF_CHANNEL: _RF_CHANNEL_PAID,
        },
        data=np.array([[[51.0, 52.0], [53.0, 54.0]]]),
        name=c.REACH,
    ),
    frequency=xr.DataArray(
        coords={
            c.GEO: ['national_geo'],
            c.MEDIA_TIME: _TIME_STRS,
            c.RF_CHANNEL: _RF_CHANNEL_PAID,
        },
        data=np.array([[[1.1, 1.2], [2, 3]]]),
        name=c.FREQUENCY,
    ),
    rf_spend=xr.DataArray(
        coords={c.RF_CHANNEL: _RF_CHANNEL_PAID},
        data=np.array([502, 504]),
        name=c.RF_SPEND,
    ),
    rf_spend_has_geo_dimension=False,
    rf_spend_has_time_dimension=False,
    kpi=xr.DataArray(
        coords={
            c.GEO: ['national_geo'],
            c.TIME: _TIME_STRS,
        },
        data=np.array([[1, 2]]),
        name=c.KPI,
    ),
    revenue_per_kpi=xr.DataArray(
        coords={
            c.GEO: ['national_geo'],
            c.TIME: _TIME_STRS,
        },
        data=np.array([[11, 12]]),
        name=c.REVENUE_PER_KPI,
    ),
    controls=xr.DataArray(
        coords={
            c.GEO: ['national_geo'],
            c.TIME: _TIME_STRS,
            c.CONTROL_VARIABLE: ['control_0', 'control_1'],
        },
        data=np.array([[[31, 32], [33, 34]]]),
        name=c.CONTROLS,
    ),
    organic_media=None,
    organic_reach=None,
    organic_frequency=None,
    non_media_treatments=None,
)

MOCK_PROTO_NATIONAL_MEDIA_RF_NON_REVENUE = text_format.Parse(
    """
    marketing_data_points {
      geo_info {
        geo_id: "national_geo"
        population: 1
      }
      date_interval {
        start_date {
          year: 2021
          month: 2
          day: 1
        }
        end_date {
          year: 2021
          month: 2
          day: 8
        }
      }
      control_variables {
        name: "control_0"
        value: 31.0
      }
      control_variables {
        name: "control_1"
        value: 32.0
      }
      media_variables {
        channel_name: "ch_paid_0"
        scalar_metric {
          name: "impressions"
          value: 41.0
        }
        media_spend: 141.0
      }
      media_variables {
        channel_name: "ch_paid_1"
        scalar_metric {
          name: "impressions"
          value: 42.0
        }
        media_spend: 142.0
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_0"
        reach: 51
        average_frequency: 1.1
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_1"
        reach: 52
        average_frequency: 1.2
      }
      kpi {
        name: "non_revenue"
        non_revenue {
          value: 1.0
          revenue_per_kpi: 11.0
        }
      }
    }
    marketing_data_points {
      geo_info {
        geo_id: "national_geo"
        population: 1
      }
      date_interval {
        start_date {
          year: 2021
          month: 2
          day: 8
        }
        end_date {
          year: 2021
          month: 2
          day: 15
        }
      }
      control_variables {
        name: "control_0"
        value: 33.0
      }
      control_variables {
        name: "control_1"
        value: 34.0
      }
      media_variables {
        channel_name: "ch_paid_0"
        scalar_metric {
          name: "impressions"
          value: 43.0
        }
        media_spend: 143.0
      }
      media_variables {
        channel_name: "ch_paid_1"
        scalar_metric {
          name: "impressions"
          value: 44.0
        }
        media_spend: 144.0
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_0"
        reach: 53
        average_frequency: 2.0
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_1"
        reach: 54
        average_frequency: 3.0
      }
      kpi {
        name: "non_revenue"
        non_revenue {
          value: 2.0
          revenue_per_kpi: 12.0
        }
      }
    }
    marketing_data_points {
      date_interval {
        start_date {
          year: 2021
          month: 2
          day: 1
        }
        end_date {
          year: 2021
          month: 2
          day: 15
        }
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_0"
        spend: 502.0
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_1"
        spend: 504.0
      }
    }
    metadata {
      time_dimensions {
        name: "time"
        dates {
          year: 2021
          month: 2
          day: 1
        }
        dates {
          year: 2021
          month: 2
          day: 8
        }
      }
      time_dimensions {
        name: "media_time"
        dates {
          year: 2021
          month: 2
          day: 1
        }
        dates {
          year: 2021
          month: 2
          day: 8
        }
      }
      channel_dimensions {
        name: "media"
        channels: "ch_paid_0"
        channels: "ch_paid_1"
      }
      channel_dimensions {
        name: "reach"
        channels: "rf_ch_paid_0"
        channels: "rf_ch_paid_1"
      }
      channel_dimensions {
        name: "frequency"
        channels: "rf_ch_paid_0"
        channels: "rf_ch_paid_1"
      }
      control_names: "control_0"
      control_names: "control_1"
      kpi_type: "non_revenue"
    }
        """,
    marketing_pb.MarketingData(),
)

# Same as above, but with no controls.
MOCK_INPUT_DATA_NATIONAL_MEDIA_RF_NON_REVENUE_NO_CONTROLS = mock.MagicMock(
    kpi_type=c.NON_REVENUE,
    geo=xr.DataArray(np.array(['national_geo'])),
    time=xr.DataArray(np.array(_TIME_STRS)),
    media_time=xr.DataArray(np.array(_TIME_STRS)),
    population=xr.DataArray(
        coords={c.GEO: ['national_geo']},
        data=np.array([1.0]),
        name=c.POPULATION,
    ),
    media=xr.DataArray(
        coords={
            c.GEO: ['national_geo'],
            c.MEDIA_TIME: _TIME_STRS,
            c.MEDIA_CHANNEL: _MEDIA_CHANNEL_PAID,
        },
        data=np.array([[[41, 42], [43, 44]]]),
        name=c.MEDIA,
    ),
    media_spend=xr.DataArray(
        coords={
            c.GEO: ['national_geo'],
            c.TIME: _TIME_STRS,
            c.MEDIA_CHANNEL: _MEDIA_CHANNEL_PAID,
        },
        data=np.array([[[141, 142], [143, 144]]]),
        name=c.MEDIA_SPEND,
    ),
    media_spend_has_geo_dimension=True,
    media_spend_has_time_dimension=True,
    reach=xr.DataArray(
        coords={
            c.GEO: ['national_geo'],
            c.MEDIA_TIME: _TIME_STRS,
            c.RF_CHANNEL: _RF_CHANNEL_PAID,
        },
        data=np.array([[[51.0, 52.0], [53.0, 54.0]]]),
        name=c.REACH,
    ),
    frequency=xr.DataArray(
        coords={
            c.GEO: ['national_geo'],
            c.MEDIA_TIME: _TIME_STRS,
            c.RF_CHANNEL: _RF_CHANNEL_PAID,
        },
        data=np.array([[[1.1, 1.2], [2, 3]]]),
        name=c.FREQUENCY,
    ),
    rf_spend=xr.DataArray(
        coords={c.RF_CHANNEL: _RF_CHANNEL_PAID},
        data=np.array([502, 504]),
        name=c.RF_SPEND,
    ),
    rf_spend_has_geo_dimension=False,
    rf_spend_has_time_dimension=False,
    kpi=xr.DataArray(
        coords={
            c.GEO: ['national_geo'],
            c.TIME: _TIME_STRS,
        },
        data=np.array([[1, 2]]),
        name=c.KPI,
    ),
    revenue_per_kpi=xr.DataArray(
        coords={
            c.GEO: ['national_geo'],
            c.TIME: _TIME_STRS,
        },
        data=np.array([[11, 12]]),
        name=c.REVENUE_PER_KPI,
    ),
    controls=None,
    organic_media=None,
    organic_reach=None,
    organic_frequency=None,
    non_media_treatments=None,
)

MOCK_PROTO_NATIONAL_MEDIA_RF_NON_REVENUE_NO_CONTROLS = text_format.Parse(
    """
    marketing_data_points {
      geo_info {
        geo_id: "national_geo"
        population: 1
      }
      date_interval {
        start_date {
          year: 2021
          month: 2
          day: 1
        }
        end_date {
          year: 2021
          month: 2
          day: 8
        }
      }
      media_variables {
        channel_name: "ch_paid_0"
        scalar_metric {
          name: "impressions"
          value: 41.0
        }
        media_spend: 141.0
      }
      media_variables {
        channel_name: "ch_paid_1"
        scalar_metric {
          name: "impressions"
          value: 42.0
        }
        media_spend: 142.0
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_0"
        reach: 51
        average_frequency: 1.1
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_1"
        reach: 52
        average_frequency: 1.2
      }
      kpi {
        name: "non_revenue"
        non_revenue {
          value: 1.0
          revenue_per_kpi: 11.0
        }
      }
    }
    marketing_data_points {
      geo_info {
        geo_id: "national_geo"
        population: 1
      }
      date_interval {
        start_date {
          year: 2021
          month: 2
          day: 8
        }
        end_date {
          year: 2021
          month: 2
          day: 15
        }
      }
      media_variables {
        channel_name: "ch_paid_0"
        scalar_metric {
          name: "impressions"
          value: 43.0
        }
        media_spend: 143.0
      }
      media_variables {
        channel_name: "ch_paid_1"
        scalar_metric {
          name: "impressions"
          value: 44.0
        }
        media_spend: 144.0
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_0"
        reach: 53
        average_frequency: 2.0
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_1"
        reach: 54
        average_frequency: 3.0
      }
      kpi {
        name: "non_revenue"
        non_revenue {
          value: 2.0
          revenue_per_kpi: 12.0
        }
      }
    }
    marketing_data_points {
      date_interval {
        start_date {
          year: 2021
          month: 2
          day: 1
        }
        end_date {
          year: 2021
          month: 2
          day: 15
        }
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_0"
        spend: 502.0
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_1"
        spend: 504.0
      }
    }
    metadata {
      time_dimensions {
        name: "time"
        dates {
          year: 2021
          month: 2
          day: 1
        }
        dates {
          year: 2021
          month: 2
          day: 8
        }
      }
      time_dimensions {
        name: "media_time"
        dates {
          year: 2021
          month: 2
          day: 1
        }
        dates {
          year: 2021
          month: 2
          day: 8
        }
      }
      channel_dimensions {
        name: "media"
        channels: "ch_paid_0"
        channels: "ch_paid_1"
      }
      channel_dimensions {
        name: "reach"
        channels: "rf_ch_paid_0"
        channels: "rf_ch_paid_1"
      }
      channel_dimensions {
        name: "frequency"
        channels: "rf_ch_paid_0"
        channels: "rf_ch_paid_1"
      }
      kpi_type: "non_revenue"
    }
        """,
    marketing_pb.MarketingData(),
)

# Media, Paid, Expanded, Lagged
MOCK_INPUT_DATA_MEDIA_PAID_EXPANDED_LAGGED = mock.MagicMock(
    kpi_type=c.REVENUE,
    geo=xr.DataArray(np.array(_GEO_IDS)),
    time=xr.DataArray(np.array(_TIME_STRS)),
    media_time=xr.DataArray(np.array(_MEDIA_TIME_STRS)),
    population=xr.DataArray(
        coords={c.GEO: _GEO_IDS},
        data=np.array([11.1, 12.2]),
        name=c.POPULATION,
    ),
    media=xr.DataArray(
        coords={
            c.GEO: _GEO_IDS,
            c.MEDIA_TIME: _MEDIA_TIME_STRS,
            c.MEDIA_CHANNEL: _MEDIA_CHANNEL_PAID,
        },
        data=np.array(
            [[[39, 40], [41, 42], [43, 44]], [[45, 46], [47, 48], [49, 50]]]
        ),
        name=c.MEDIA,
    ),
    media_spend=xr.DataArray(
        coords={c.MEDIA_CHANNEL: _MEDIA_CHANNEL_PAID},
        data=np.array([492, 496]),
        name=c.MEDIA_SPEND,
    ),
    media_spend_has_geo_dimension=False,
    media_spend_has_time_dimension=False,
    reach=None,
    frequency=None,
    rf_spend=None,
    kpi=xr.DataArray(
        coords={
            c.GEO: _GEO_IDS,
            c.TIME: _TIME_STRS,
        },
        data=np.array([[2, 3], [4, 5]]),
        name=c.KPI,
    ),
    revenue_per_kpi=xr.DataArray(
        coords={
            c.GEO: _GEO_IDS,
            c.TIME: _TIME_STRS,
        },
        data=np.ones((2, 2)),
        name=c.REVENUE_PER_KPI,
    ),
    controls=xr.DataArray(
        coords={
            c.GEO: _GEO_IDS,
            c.TIME: _TIME_STRS,
            c.CONTROL_VARIABLE: _CONTROL_VARIABLES,
        },
        data=np.array([[[32, 33], [34, 35]], [[36, 37], [38, 39]]]),
        name=c.CONTROLS,
    ),
    organic_media=None,
    organic_reach=None,
    organic_frequency=None,
    non_media_treatments=None,
)

MOCK_PROTO_MEDIA_PAID_EXPANDED_LAGGED = text_format.Parse(
    """
    marketing_data_points {
      geo_info {
        geo_id: "geo_0"
        population: 11
      }
      date_interval {
        start_date {
          year: 2021
          month: 1
          day: 25
        }
        end_date {
          year: 2021
          month: 2
          day: 1
        }
      }
      media_variables {
        channel_name: "ch_paid_0"
        scalar_metric {
          name: "impressions"
          value: 39.0
        }
      }
      media_variables {
        channel_name: "ch_paid_1"
        scalar_metric {
          name: "impressions"
          value: 40.0
        }
      }
    }
    marketing_data_points {
      geo_info {
        geo_id: "geo_0"
        population: 11
      }
      date_interval {
        start_date {
          year: 2021
          month: 2
          day: 1
        }
        end_date {
          year: 2021
          month: 2
          day: 8
        }
      }
      control_variables {
        name: "control_0"
        value: 32.0
      }
      control_variables {
        name: "control_1"
        value: 33.0
      }
      media_variables {
        channel_name: "ch_paid_0"
        scalar_metric {
          name: "impressions"
          value: 41.0
        }
      }
      media_variables {
        channel_name: "ch_paid_1"
        scalar_metric {
          name: "impressions"
          value: 42.0
        }
      }
      kpi {
        name: "revenue"
        revenue {
          value: 2.0
        }
      }
    }
    marketing_data_points {
      geo_info {
        geo_id: "geo_0"
        population: 11
      }
      date_interval {
        start_date {
          year: 2021
          month: 2
          day: 8
        }
        end_date {
          year: 2021
          month: 2
          day: 15
        }
      }
      control_variables {
        name: "control_0"
        value: 34.0
      }
      control_variables {
        name: "control_1"
        value: 35.0
      }
      media_variables {
        channel_name: "ch_paid_0"
        scalar_metric {
          name: "impressions"
          value: 43.0
        }
      }
      media_variables {
        channel_name: "ch_paid_1"
        scalar_metric {
          name: "impressions"
          value: 44.0
        }
      }
      kpi {
        name: "revenue"
        revenue {
          value: 3.0
        }
      }
    }
    marketing_data_points {
      geo_info {
        geo_id: "geo_1"
        population: 12
      }
      date_interval {
        start_date {
          year: 2021
          month: 1
          day: 25
        }
        end_date {
          year: 2021
          month: 2
          day: 1
        }
      }
      media_variables {
        channel_name: "ch_paid_0"
        scalar_metric {
          name: "impressions"
          value: 45.0
        }
      }
      media_variables {
        channel_name: "ch_paid_1"
        scalar_metric {
          name: "impressions"
          value: 46.0
        }
      }
    }
    marketing_data_points {
      geo_info {
        geo_id: "geo_1"
        population: 12
      }
      date_interval {
        start_date {
          year: 2021
          month: 2
          day: 1
        }
        end_date {
          year: 2021
          month: 2
          day: 8
        }
      }
      control_variables {
        name: "control_0"
        value: 36.0
      }
      control_variables {
        name: "control_1"
        value: 37.0
      }
      media_variables {
        channel_name: "ch_paid_0"
        scalar_metric {
          name: "impressions"
          value: 47.0
        }
      }
      media_variables {
        channel_name: "ch_paid_1"
        scalar_metric {
          name: "impressions"
          value: 48.0
        }
      }
      kpi {
        name: "revenue"
        revenue {
          value: 4.0
        }
      }
    }
    marketing_data_points {
      geo_info {
        geo_id: "geo_1"
        population: 12
      }
      date_interval {
        start_date {
          year: 2021
          month: 2
          day: 8
        }
        end_date {
          year: 2021
          month: 2
          day: 15
        }
      }
      control_variables {
        name: "control_0"
        value: 38.0
      }
      control_variables {
        name: "control_1"
        value: 39.0
      }
      media_variables {
        channel_name: "ch_paid_0"
        scalar_metric {
          name: "impressions"
          value: 49.0
        }
      }
      media_variables {
        channel_name: "ch_paid_1"
        scalar_metric {
          name: "impressions"
          value: 50.0
        }
      }
      kpi {
        name: "revenue"
        revenue {
          value: 5.0
        }
      }
    }
    marketing_data_points {
      date_interval {
        start_date {
          year: 2021
          month: 1
          day: 25
        }
        end_date {
          year: 2021
          month: 2
          day: 15
        }
      }
      media_variables {
        channel_name: "ch_paid_0"
        media_spend: 492.0
      }
      media_variables {
        channel_name: "ch_paid_1"
        media_spend: 496.0
      }
    }
    metadata {
      time_dimensions {
        name: "time"
        dates {
          year: 2021
          month: 2
          day: 1
        }
        dates {
          year: 2021
          month: 2
          day: 8
        }
      }
      time_dimensions {
        name: "media_time"
        dates {
          year: 2021
          month: 1
          day: 25
        }
        dates {
          year: 2021
          month: 2
          day: 1
        }
        dates {
          year: 2021
          month: 2
          day: 8
        }
      }
      channel_dimensions {
        name: "media"
        channels: "ch_paid_0"
        channels: "ch_paid_1"
      }
      control_names: "control_0"
      control_names: "control_1"
      kpi_type: "revenue"
    }
    """,
    marketing_pb.MarketingData(),
)

# Media, Paid, Granular, Not Lagged
MOCK_INPUT_DATA_MEDIA_PAID_GRANULAR_NOT_LAGGED = mock.MagicMock(
    kpi_type=c.REVENUE,
    geo=xr.DataArray(np.array(_GEO_IDS)),
    time=xr.DataArray(np.array(_TIME_STRS)),
    media_time=xr.DataArray(np.array(_TIME_STRS)),
    population=xr.DataArray(
        coords={c.GEO: _GEO_IDS}, data=np.array([11.1, 12.2]), name=c.POPULATION
    ),
    media=xr.DataArray(
        coords={
            c.GEO: _GEO_IDS,
            c.MEDIA_TIME: _TIME_STRS,
            c.MEDIA_CHANNEL: _MEDIA_CHANNEL_PAID,
        },
        data=np.array([[[39, 40], [41, 42]], [[43, 44], [45, 46]]]),
        name=c.MEDIA,
    ),
    media_spend=xr.DataArray(
        coords={
            c.GEO: _GEO_IDS,
            c.TIME: _TIME_STRS,
            c.MEDIA_CHANNEL: _MEDIA_CHANNEL_PAID,
        },
        data=np.array([[[123, 124], [125, 126]], [[127, 128], [129, 130]]]),
        name=c.MEDIA_SPEND,
    ),
    media_spend_has_geo_dimension=True,
    media_spend_has_time_dimension=True,
    reach=None,
    frequency=None,
    rf_spend=None,
    kpi=xr.DataArray(
        coords={c.GEO: _GEO_IDS, c.TIME: _TIME_STRS},
        data=np.array([[1, 2], [3, 4]]),
        name=c.KPI,
    ),
    revenue_per_kpi=xr.DataArray(
        coords={c.GEO: _GEO_IDS, c.TIME: _TIME_STRS},
        data=np.ones((2, 2)),
        name=c.REVENUE_PER_KPI,
    ),
    controls=xr.DataArray(
        coords={
            c.GEO: _GEO_IDS,
            c.TIME: _TIME_STRS,
            c.CONTROL_VARIABLE: _CONTROL_VARIABLES,
        },
        data=np.array([[[31, 32], [33, 34]], [[35, 36], [37, 38]]]),
        name=c.CONTROLS,
    ),
    organic_media=None,
    organic_reach=None,
    organic_frequency=None,
    non_media_treatments=None,
)

MOCK_PROTO_MEDIA_PAID_GRANULAR_NOT_LAGGED_STRING = """
   marketing_data_points {
  geo_info {
    geo_id: "geo_0"
    population: 11
  }
  date_interval {
    start_date {
      year: 2021
      month: 2
      day: 1
    }
    end_date {
      year: 2021
      month: 2
      day: 8
    }
  }
  control_variables {
    name: "control_0"
    value: 31.0
  }
  control_variables {
    name: "control_1"
    value: 32.0
  }
  media_variables {
    channel_name: "ch_paid_0"
    scalar_metric {
      name: "impressions"
      value: 39.0
    }
    media_spend: 123.0
  }
  media_variables {
    channel_name: "ch_paid_1"
    scalar_metric {
      name: "impressions"
      value: 40.0
    }
    media_spend: 124.0
  }
  kpi {
    name: "revenue"
    revenue {
      value: 1.0
    }
  }
}
marketing_data_points {
  geo_info {
    geo_id: "geo_0"
    population: 11
  }
  date_interval {
    start_date {
      year: 2021
      month: 2
      day: 8
    }
    end_date {
      year: 2021
      month: 2
      day: 15
    }
  }
  control_variables {
    name: "control_0"
    value: 33.0
  }
  control_variables {
    name: "control_1"
    value: 34.0
  }
  media_variables {
    channel_name: "ch_paid_0"
    scalar_metric {
      name: "impressions"
      value: 41.0
    }
    media_spend: 125.0
  }
  media_variables {
    channel_name: "ch_paid_1"
    scalar_metric {
      name: "impressions"
      value: 42.0
    }
    media_spend: 126.0
  }
  kpi {
    name: "revenue"
    revenue {
      value: 2.0
    }
  }
}
marketing_data_points {
  geo_info {
    geo_id: "geo_1"
    population: 12
  }
  date_interval {
    start_date {
      year: 2021
      month: 2
      day: 1
    }
    end_date {
      year: 2021
      month: 2
      day: 8
    }
  }
  control_variables {
    name: "control_0"
    value: 35.0
  }
  control_variables {
    name: "control_1"
    value: 36.0
  }
  media_variables {
    channel_name: "ch_paid_0"
    scalar_metric {
      name: "impressions"
      value: 43.0
    }
    media_spend: 127.0
  }
  media_variables {
    channel_name: "ch_paid_1"
    scalar_metric {
      name: "impressions"
      value: 44.0
    }
    media_spend: 128.0
  }
  kpi {
    name: "revenue"
    revenue {
      value: 3.0
    }
  }
}
marketing_data_points {
  geo_info {
    geo_id: "geo_1"
    population: 12
  }
  date_interval {
    start_date {
      year: 2021
      month: 2
      day: 8
    }
    end_date {
      year: 2021
      month: 2
      day: 15
    }
  }
  control_variables {
    name: "control_0"
    value: 37.0
  }
  control_variables {
    name: "control_1"
    value: 38.0
  }
  media_variables {
    channel_name: "ch_paid_0"
    scalar_metric {
      name: "impressions"
      value: 45.0
    }
    media_spend: 129.0
  }
  media_variables {
    channel_name: "ch_paid_1"
    scalar_metric {
      name: "impressions"
      value: 46.0
    }
    media_spend: 130.0
  }
  kpi {
    name: "revenue"
    revenue {
      value: 4.0
    }
  }
}
metadata {
  time_dimensions {
    name: "time"
    dates {
      year: 2021
      month: 2
      day: 1
    }
    dates {
      year: 2021
      month: 2
      day: 8
    }
  }
  time_dimensions {
    name: "media_time"
    dates {
      year: 2021
      month: 2
      day: 1
    }
    dates {
      year: 2021
      month: 2
      day: 8
    }
  }
  channel_dimensions {
    name: "media"
    channels: "ch_paid_0"
    channels: "ch_paid_1"
  }
  control_names: "control_0"
  control_names: "control_1"
  kpi_type: "revenue"
}
"""

MOCK_PROTO_MEDIA_PAID_GRANULAR_NOT_LAGGED = text_format.Parse(
    MOCK_PROTO_MEDIA_PAID_GRANULAR_NOT_LAGGED_STRING,
    marketing_pb.MarketingData(),
)

# Media, Organic, Expanded, Lagged
MOCK_INPUT_DATA_MEDIA_ORGANIC_EXPANDED_LAGGED = mock.MagicMock(
    kpi_type=c.REVENUE,
    geo=xr.DataArray(np.array(_GEO_IDS)),
    time=xr.DataArray(np.array(_TIME_STRS)),
    media_time=xr.DataArray(np.array(_MEDIA_TIME_STRS)),
    population=xr.DataArray(
        coords={c.GEO: _GEO_IDS}, data=np.array([11.1, 12.2]), name=c.POPULATION
    ),
    media=xr.DataArray(
        coords={
            c.GEO: _GEO_IDS,
            c.MEDIA_TIME: _MEDIA_TIME_STRS,
            c.MEDIA_CHANNEL: _MEDIA_CHANNEL_PAID,
        },
        data=np.array([[[1, 2], [3, 4], [5, 6]], [[7, 8], [9, 10], [11, 12]]]),
        name=c.MEDIA,
    ),
    media_spend=xr.DataArray(
        coords={c.MEDIA_CHANNEL: _MEDIA_CHANNEL_PAID},
        data=np.array([492, 496]),
        name=c.MEDIA_SPEND,
    ),
    media_spend_has_geo_dimension=False,
    media_spend_has_time_dimension=False,
    reach=None,
    frequency=None,
    rf_spend=None,
    organic_media=xr.DataArray(
        coords={
            c.GEO: _GEO_IDS,
            c.MEDIA_TIME: _MEDIA_TIME_STRS,
            c.ORGANIC_MEDIA_CHANNEL: _MEDIA_CHANNEL_ORGANIC,
        },
        data=np.array(
            [[[39, 40], [41, 42], [43, 44]], [[45, 46], [47, 48], [49, 50]]]
        ),
        name=c.ORGANIC_MEDIA,
    ),
    organic_reach=None,
    organic_frequency=None,
    non_media_treatments=None,
    kpi=xr.DataArray(
        coords={c.GEO: _GEO_IDS, c.TIME: _TIME_STRS},
        data=np.array([[2, 2], [3, 3]]),
        name=c.KPI,
    ),
    revenue_per_kpi=xr.DataArray(
        coords={c.GEO: _GEO_IDS, c.TIME: _TIME_STRS},
        data=np.ones((2, 2)),
        name=c.REVENUE_PER_KPI,
    ),
    controls=xr.DataArray(
        coords={
            c.GEO: _GEO_IDS,
            c.TIME: _TIME_STRS,
            c.CONTROL_VARIABLE: _CONTROL_VARIABLES,
        },
        data=np.array([[[31, 32], [33, 34]], [[35, 36], [37, 38]]]),
        name=c.CONTROLS,
    ),
)

MOCK_PROTO_MEDIA_ORGANIC_EXPANDED_LAGGED = text_format.Parse(
    """
    marketing_data_points {
      geo_info {
        geo_id: "geo_0"
        population: 11
      }
      date_interval {
        start_date {
          year: 2021
          month: 1
          day: 25
        }
        end_date {
          year: 2021
          month: 2
          day: 1
        }
      }
      media_variables {
        channel_name: "ch_paid_0"
        scalar_metric {
          name: "impressions"
          value: 1.0
        }
      }
      media_variables {
        channel_name: "ch_paid_1"
        scalar_metric {
          name: "impressions"
          value: 2.0
        }
      }
      media_variables {
        channel_name: "ch_organic_0"
        scalar_metric {
          name: "impressions"
          value: 39.0
        }
      }
      media_variables {
        channel_name: "ch_organic_1"
        scalar_metric {
          name: "impressions"
          value: 40.0
        }
      }
    }
    marketing_data_points {
      geo_info {
        geo_id: "geo_0"
        population: 11
      }
      date_interval {
        start_date {
          year: 2021
          month: 2
          day: 1
        }
        end_date {
          year: 2021
          month: 2
          day: 8
        }
      }
      control_variables {
        name: "control_0"
        value: 31.0
      }
      control_variables {
        name: "control_1"
        value: 32.0
      }
      media_variables {
        channel_name: "ch_paid_0"
        scalar_metric {
          name: "impressions"
          value: 3.0
        }
      }
      media_variables {
        channel_name: "ch_paid_1"
        scalar_metric {
          name: "impressions"
          value: 4.0
        }
      }
      media_variables {
        channel_name: "ch_organic_0"
        scalar_metric {
          name: "impressions"
          value: 41.0
        }
      }
      media_variables {
        channel_name: "ch_organic_1"
        scalar_metric {
          name: "impressions"
          value: 42.0
        }
      }
      kpi {
        name: "revenue"
        revenue {
          value: 2.0
        }
      }
    }
    marketing_data_points {
      geo_info {
        geo_id: "geo_0"
        population: 11
      }
      date_interval {
        start_date {
          year: 2021
          month: 2
          day: 8
        }
        end_date {
          year: 2021
          month: 2
          day: 15
        }
      }
      control_variables {
        name: "control_0"
        value: 33.0
      }
      control_variables {
        name: "control_1"
        value: 34.0
      }
      media_variables {
        channel_name: "ch_paid_0"
        scalar_metric {
          name: "impressions"
          value: 5.0
        }
      }
      media_variables {
        channel_name: "ch_paid_1"
        scalar_metric {
          name: "impressions"
          value: 6.0
        }
      }
      media_variables {
        channel_name: "ch_organic_0"
        scalar_metric {
          name: "impressions"
          value: 43.0
        }
      }
      media_variables {
        channel_name: "ch_organic_1"
        scalar_metric {
          name: "impressions"
          value: 44.0
        }
      }
      kpi {
        name: "revenue"
        revenue {
          value: 2.0
        }
      }
    }
    marketing_data_points {
      geo_info {
        geo_id: "geo_1"
        population: 12
      }
      date_interval {
        start_date {
          year: 2021
          month: 1
          day: 25
        }
        end_date {
          year: 2021
          month: 2
          day: 1
        }
      }
      media_variables {
        channel_name: "ch_paid_0"
        scalar_metric {
          name: "impressions"
          value: 7.0
        }
      }
      media_variables {
        channel_name: "ch_paid_1"
        scalar_metric {
          name: "impressions"
          value: 8.0
        }
      }
      media_variables {
        channel_name: "ch_organic_0"
        scalar_metric {
          name: "impressions"
          value: 45.0
        }
      }
      media_variables {
        channel_name: "ch_organic_1"
        scalar_metric {
          name: "impressions"
          value: 46.0
        }
      }
    }
    marketing_data_points {
      geo_info {
        geo_id: "geo_1"
        population: 12
      }
      date_interval {
        start_date {
          year: 2021
          month: 2
          day: 1
        }
        end_date {
          year: 2021
          month: 2
          day: 8
        }
      }
      control_variables {
        name: "control_0"
        value: 35.0
      }
      control_variables {
        name: "control_1"
        value: 36.0
      }
      media_variables {
        channel_name: "ch_paid_0"
        scalar_metric {
          name: "impressions"
          value: 9.0
        }
      }
      media_variables {
        channel_name: "ch_paid_1"
        scalar_metric {
          name: "impressions"
          value: 10.0
        }
      }
      media_variables {
        channel_name: "ch_organic_0"
        scalar_metric {
          name: "impressions"
          value: 47.0
        }
      }
      media_variables {
        channel_name: "ch_organic_1"
        scalar_metric {
          name: "impressions"
          value: 48.0
        }
      }
      kpi {
        name: "revenue"
        revenue {
          value: 3.0
        }
      }
    }
    marketing_data_points {
      geo_info {
        geo_id: "geo_1"
        population: 12
      }
      date_interval {
        start_date {
          year: 2021
          month: 2
          day: 8
        }
        end_date {
          year: 2021
          month: 2
          day: 15
        }
      }
      control_variables {
        name: "control_0"
        value: 37.0
      }
      control_variables {
        name: "control_1"
        value: 38.0
      }
      media_variables {
        channel_name: "ch_paid_0"
        scalar_metric {
          name: "impressions"
          value: 11.0
        }
      }
      media_variables {
        channel_name: "ch_paid_1"
        scalar_metric {
          name: "impressions"
          value: 12.0
        }
      }
      media_variables {
        channel_name: "ch_organic_0"
        scalar_metric {
          name: "impressions"
          value: 49.0
        }
      }
      media_variables {
        channel_name: "ch_organic_1"
        scalar_metric {
          name: "impressions"
          value: 50.0
        }
      }
      kpi {
        name: "revenue"
        revenue {
          value: 3.0
        }
      }
    }
    marketing_data_points {
      date_interval {
        start_date {
          year: 2021
          month: 1
          day: 25
        }
        end_date {
          year: 2021
          month: 2
          day: 15
        }
      }
      media_variables {
        channel_name: "ch_paid_0"
        media_spend: 492.0
      }
      media_variables {
        channel_name: "ch_paid_1"
        media_spend: 496.0
      }
    }
    metadata {
      time_dimensions {
        name: "time"
        dates {
          year: 2021
          month: 2
          day: 1
        }
        dates {
          year: 2021
          month: 2
          day: 8
        }
      }
      time_dimensions {
        name: "media_time"
        dates {
          year: 2021
          month: 1
          day: 25
        }
        dates {
          year: 2021
          month: 2
          day: 1
        }
        dates {
          year: 2021
          month: 2
          day: 8
        }
      }
      channel_dimensions {
        name: "media"
        channels: "ch_paid_0"
        channels: "ch_paid_1"
      }
      channel_dimensions {
        name: "organic_media"
        channels: "ch_organic_0"
        channels: "ch_organic_1"
      }
      control_names: "control_0"
      control_names: "control_1"
      kpi_type: "revenue"
    }
    """,
    marketing_pb.MarketingData(),
)

# Media, Organic, Granular, Not Lagged
MOCK_INPUT_DATA_MEDIA_ORGANIC_GRANULAR_NOT_LAGGED = mock.MagicMock(
    kpi_type=c.REVENUE,
    geo=xr.DataArray(np.array(_GEO_IDS)),
    time=xr.DataArray(np.array(_TIME_STRS)),
    media_time=xr.DataArray(np.array(_TIME_STRS)),
    population=xr.DataArray(
        coords={c.GEO: _GEO_IDS}, data=np.array([11.1, 12.2]), name=c.POPULATION
    ),
    media=xr.DataArray(
        coords={
            c.GEO: _GEO_IDS,
            c.MEDIA_TIME: _TIME_STRS,
            c.MEDIA_CHANNEL: _MEDIA_CHANNEL_PAID,
        },
        data=np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]]),
        name=c.MEDIA,
    ),
    media_spend=xr.DataArray(
        coords={
            c.GEO: _GEO_IDS,
            c.TIME: _TIME_STRS,
            c.MEDIA_CHANNEL: _MEDIA_CHANNEL_PAID,
        },
        data=np.array([[[123, 124], [125, 126]], [[127, 128], [129, 130]]]),
        name=c.MEDIA_SPEND,
    ),
    media_spend_has_geo_dimension=True,
    media_spend_has_time_dimension=True,
    reach=None,
    frequency=None,
    rf_spend=None,
    organic_media=xr.DataArray(
        coords={
            c.GEO: _GEO_IDS,
            c.MEDIA_TIME: _TIME_STRS,
            c.ORGANIC_MEDIA_CHANNEL: _MEDIA_CHANNEL_ORGANIC,
        },
        data=np.array([[[39, 40], [41, 42]], [[43, 44], [45, 46]]]),
        name=c.ORGANIC_MEDIA,
    ),
    organic_reach=None,
    organic_frequency=None,
    non_media_treatments=None,
    kpi=xr.DataArray(
        coords={c.GEO: _GEO_IDS, c.TIME: _TIME_STRS},
        data=np.array([[2, 2], [3, 3]]),
        name=c.KPI,
    ),
    revenue_per_kpi=xr.DataArray(
        coords={c.GEO: _GEO_IDS, c.TIME: _TIME_STRS},
        data=np.ones((2, 2)),
        name=c.REVENUE_PER_KPI,
    ),
    controls=xr.DataArray(
        coords={
            c.GEO: _GEO_IDS,
            c.TIME: _TIME_STRS,
            c.CONTROL_VARIABLE: _CONTROL_VARIABLES,
        },
        data=np.array([[[31, 32], [33, 34]], [[35, 36], [37, 38]]]),
        name=c.CONTROLS,
    ),
)

MOCK_PROTO_MEDIA_ORGANIC_GRANULAR_NOT_LAGGED = text_format.Parse(
    """
    marketing_data_points {
      geo_info {
        geo_id: "geo_0"
        population: 11
      }
      date_interval {
        start_date {
          year: 2021
          month: 2
          day: 1
        }
        end_date {
          year: 2021
          month: 2
          day: 8
        }
      }
      control_variables {
        name: "control_0"
        value: 31.0
      }
      control_variables {
        name: "control_1"
        value: 32.0
      }
      media_variables {
        channel_name: "ch_paid_0"
        scalar_metric {
          name: "impressions"
          value: 1.0
        }
        media_spend: 123.0
      }
      media_variables {
        channel_name: "ch_paid_1"
        scalar_metric {
          name: "impressions"
          value: 2.0
        }
        media_spend: 124.0
      }
      media_variables {
        channel_name: "ch_organic_0"
        scalar_metric {
          name: "impressions"
          value: 39.0
        }
      }
      media_variables {
        channel_name: "ch_organic_1"
        scalar_metric {
          name: "impressions"
          value: 40.0
        }
      }
      kpi {
        name: "revenue"
        revenue {
          value: 2.0
        }
      }
    }
    marketing_data_points {
      geo_info {
        geo_id: "geo_0"
        population: 11
      }
      date_interval {
        start_date {
          year: 2021
          month: 2
          day: 8
        }
        end_date {
          year: 2021
          month: 2
          day: 15
        }
      }
      control_variables {
        name: "control_0"
        value: 33.0
      }
      control_variables {
        name: "control_1"
        value: 34.0
      }
      media_variables {
        channel_name: "ch_paid_0"
        scalar_metric {
          name: "impressions"
          value: 3.0
        }
        media_spend: 125.0
      }
      media_variables {
        channel_name: "ch_paid_1"
        scalar_metric {
          name: "impressions"
          value: 4.0
        }
        media_spend: 126.0
      }
      media_variables {
        channel_name: "ch_organic_0"
        scalar_metric {
          name: "impressions"
          value: 41.0
        }
      }
      media_variables {
        channel_name: "ch_organic_1"
        scalar_metric {
          name: "impressions"
          value: 42.0
        }
      }
      kpi {
        name: "revenue"
        revenue {
          value: 2.0
        }
      }
    }
    marketing_data_points {
      geo_info {
        geo_id: "geo_1"
        population: 12
      }
      date_interval {
        start_date {
          year: 2021
          month: 2
          day: 1
        }
        end_date {
          year: 2021
          month: 2
          day: 8
        }
      }
      control_variables {
        name: "control_0"
        value: 35.0
      }
      control_variables {
        name: "control_1"
        value: 36.0
      }
      media_variables {
        channel_name: "ch_paid_0"
        scalar_metric {
          name: "impressions"
          value: 5.0
        }
        media_spend: 127.0
      }
      media_variables {
        channel_name: "ch_paid_1"
        scalar_metric {
          name: "impressions"
          value: 6.0
        }
        media_spend: 128.0
      }
      media_variables {
        channel_name: "ch_organic_0"
        scalar_metric {
          name: "impressions"
          value: 43.0
        }
      }
      media_variables {
        channel_name: "ch_organic_1"
        scalar_metric {
          name: "impressions"
          value: 44.0
        }
      }
      kpi {
        name: "revenue"
        revenue {
          value: 3.0
        }
      }
    }
    marketing_data_points {
      geo_info {
        geo_id: "geo_1"
        population: 12
      }
      date_interval {
        start_date {
          year: 2021
          month: 2
          day: 8
        }
        end_date {
          year: 2021
          month: 2
          day: 15
        }
      }
      control_variables {
        name: "control_0"
        value: 37.0
      }
      control_variables {
        name: "control_1"
        value: 38.0
      }
      media_variables {
        channel_name: "ch_paid_0"
        scalar_metric {
          name: "impressions"
          value: 7.0
        }
        media_spend: 129.0
      }
      media_variables {
        channel_name: "ch_paid_1"
        scalar_metric {
          name: "impressions"
          value: 8.0
        }
        media_spend: 130.0
      }
      media_variables {
        channel_name: "ch_organic_0"
        scalar_metric {
          name: "impressions"
          value: 45.0
        }
      }
      media_variables {
        channel_name: "ch_organic_1"
        scalar_metric {
          name: "impressions"
          value: 46.0
        }
      }
      kpi {
        name: "revenue"
        revenue {
          value: 3.0
        }
      }
    }
    metadata {
      time_dimensions {
        name: "time"
        dates {
          year: 2021
          month: 2
          day: 1
        }
        dates {
          year: 2021
          month: 2
          day: 8
        }
      }
      time_dimensions {
        name: "media_time"
        dates {
          year: 2021
          month: 2
          day: 1
        }
        dates {
          year: 2021
          month: 2
          day: 8
        }
      }
      channel_dimensions {
        name: "media"
        channels: "ch_paid_0"
        channels: "ch_paid_1"
      }
      channel_dimensions {
        name: "organic_media"
        channels: "ch_organic_0"
        channels: "ch_organic_1"
      }
      control_names: "control_0"
      control_names: "control_1"
      kpi_type: "revenue"
    }
    """,
    marketing_pb.MarketingData(),
)

# Reach and Frequency, Paid, Expanded, Lagged
MOCK_INPUT_DATA_RF_PAID_EXPANDED_LAGGED = mock.MagicMock(
    kpi_type=c.REVENUE,
    geo=xr.DataArray(np.array(_GEO_IDS)),
    time=xr.DataArray(np.array(_TIME_STRS)),
    media_time=xr.DataArray(np.array(_MEDIA_TIME_STRS)),
    population=xr.DataArray(
        coords={c.GEO: _GEO_IDS},
        data=np.array([11.1, 12.2]),
        name=c.POPULATION,
    ),
    media=None,
    media_spend=None,
    reach=xr.DataArray(
        coords={
            c.GEO: _GEO_IDS,
            c.MEDIA_TIME: _MEDIA_TIME_STRS,
            c.RF_CHANNEL: _RF_CHANNEL_PAID,
        },
        data=np.array(
            [[[51, 52], [53, 54], [55, 56]], [[57, 58], [59, 60], [61, 62]]]
        ),
        name=c.REACH,
    ),
    frequency=xr.DataArray(
        coords={
            c.GEO: _GEO_IDS,
            c.MEDIA_TIME: _MEDIA_TIME_STRS,
            c.RF_CHANNEL: _RF_CHANNEL_PAID,
        },
        data=np.array([
            [[1.1, 1.2], [1.3, 1.4], [1.5, 1.6]],
            [[1.7, 1.8], [1.9, 2.0], [2.1, 2.2]],
        ]),
        name=c.FREQUENCY,
    ),
    rf_spend=xr.DataArray(
        coords={c.RF_CHANNEL: _RF_CHANNEL_PAID},
        data=np.array([1004, 1008]),
        name=c.RF_SPEND,
    ),
    rf_spend_has_geo_dimension=False,
    rf_spend_has_time_dimension=False,
    kpi=xr.DataArray(
        coords={
            c.GEO: _GEO_IDS,
            c.TIME: _TIME_STRS,
        },
        data=np.array([[2, 3], [4, 5]]),
        name=c.KPI,
    ),
    revenue_per_kpi=xr.DataArray(
        coords={
            c.GEO: _GEO_IDS,
            c.TIME: _TIME_STRS,
        },
        data=np.ones((2, 2)),
        name=c.REVENUE_PER_KPI,
    ),
    controls=xr.DataArray(
        coords={
            c.GEO: _GEO_IDS,
            c.TIME: _TIME_STRS,
            c.CONTROL_VARIABLE: _CONTROL_VARIABLES,
        },
        data=np.array([[[32, 33], [34, 35]], [[36, 37], [38, 39]]]),
        name=c.CONTROLS,
    ),
    organic_media=None,
    organic_reach=None,
    organic_frequency=None,
    non_media_treatments=None,
)

MOCK_PROTO_RF_PAID_EXPANDED_LAGGED = text_format.Parse(
    """
    marketing_data_points {
      geo_info {
        geo_id: "geo_0"
        population: 11
      }
      date_interval {
        start_date {
          year: 2021
          month: 1
          day: 25
        }
        end_date {
          year: 2021
          month: 2
          day: 1
        }
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_0"
        reach: 51
        average_frequency: 1.1
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_1"
        reach: 52
        average_frequency: 1.2
      }
    }
    marketing_data_points {
      geo_info {
        geo_id: "geo_0"
        population: 11
      }
      date_interval {
        start_date {
          year: 2021
          month: 2
          day: 1
        }
        end_date {
          year: 2021
          month: 2
          day: 8
        }
      }
      control_variables {
        name: "control_0"
        value: 32.0
      }
      control_variables {
        name: "control_1"
        value: 33.0
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_0"
        reach: 53
        average_frequency: 1.3
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_1"
        reach: 54
        average_frequency: 1.4
      }
      kpi {
        name: "revenue"
        revenue {
          value: 2.0
        }
      }
    }
    marketing_data_points {
      geo_info {
        geo_id: "geo_0"
        population: 11
      }
      date_interval {
        start_date {
          year: 2021
          month: 2
          day: 8
        }
        end_date {
          year: 2021
          month: 2
          day: 15
        }
      }
      control_variables {
        name: "control_0"
        value: 34.0
      }
      control_variables {
        name: "control_1"
        value: 35.0
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_0"
        reach: 55
        average_frequency: 1.5
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_1"
        reach: 56
        average_frequency: 1.6
      }
      kpi {
        name: "revenue"
        revenue {
          value: 3.0
        }
      }
    }
    marketing_data_points {
      geo_info {
        geo_id: "geo_1"
        population: 12
      }
      date_interval {
        start_date {
          year: 2021
          month: 1
          day: 25
        }
        end_date {
          year: 2021
          month: 2
          day: 1
        }
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_0"
        reach: 57
        average_frequency: 1.7
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_1"
        reach: 58
        average_frequency: 1.8
      }
    }
    marketing_data_points {
      geo_info {
        geo_id: "geo_1"
        population: 12
      }
      date_interval {
        start_date {
          year: 2021
          month: 2
          day: 1
        }
        end_date {
          year: 2021
          month: 2
          day: 8
        }
      }
      control_variables {
        name: "control_0"
        value: 36.0
      }
      control_variables {
        name: "control_1"
        value: 37.0
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_0"
        reach: 59
        average_frequency: 1.9
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_1"
        reach: 60
        average_frequency: 2.0
      }
      kpi {
        name: "revenue"
        revenue {
          value: 4.0
        }
      }
    }
    marketing_data_points {
      geo_info {
        geo_id: "geo_1"
        population: 12
      }
      date_interval {
        start_date {
          year: 2021
          month: 2
          day: 8
        }
        end_date {
          year: 2021
          month: 2
          day: 15
        }
      }
      control_variables {
        name: "control_0"
        value: 38.0
      }
      control_variables {
        name: "control_1"
        value: 39.0
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_0"
        reach: 61
        average_frequency: 2.1
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_1"
        reach: 62
        average_frequency: 2.2
      }
      kpi {
        name: "revenue"
        revenue {
          value: 5.0
        }
      }
    }
    marketing_data_points {
      date_interval {
        start_date {
          year: 2021
          month: 1
          day: 25
        }
        end_date {
          year: 2021
          month: 2
          day: 15
        }
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_0"
        spend: 1004.0
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_1"
        spend: 1008.0
      }
    }
    metadata {
      time_dimensions {
        name: "time"
        dates {
          year: 2021
          month: 2
          day: 1
        }
        dates {
          year: 2021
          month: 2
          day: 8
        }
      }
      time_dimensions {
        name: "media_time"
        dates {
          year: 2021
          month: 1
          day: 25
        }
        dates {
          year: 2021
          month: 2
          day: 1
        }
        dates {
          year: 2021
          month: 2
          day: 8
        }
      }
      channel_dimensions {
        name: "reach"
        channels: "rf_ch_paid_0"
        channels: "rf_ch_paid_1"
      }
      channel_dimensions {
        name: "frequency"
        channels: "rf_ch_paid_0"
        channels: "rf_ch_paid_1"
      }
      control_names: "control_0"
      control_names: "control_1"
      kpi_type: "revenue"
    }
    """,
    marketing_pb.MarketingData(),
)

# Reach and Frequency, Paid, Granular, Not Lagged
MOCK_INPUT_DATA_RF_PAID_GRANULAR_NOT_LAGGED = mock.MagicMock(
    kpi_type=c.REVENUE,
    geo=xr.DataArray(np.array(_GEO_IDS)),
    time=xr.DataArray(np.array(_TIME_STRS)),
    media_time=xr.DataArray(np.array(_TIME_STRS)),
    population=xr.DataArray(
        coords={c.GEO: _GEO_IDS}, data=np.array([11.1, 12.2]), name=c.POPULATION
    ),
    media=None,
    media_spend=None,
    reach=xr.DataArray(
        coords={
            c.GEO: _GEO_IDS,
            c.MEDIA_TIME: _TIME_STRS,
            c.RF_CHANNEL: _RF_CHANNEL_PAID,
        },
        data=np.array(
            [[[51.0, 52.0], [53.0, 54.0]], [[55.0, 56.0], [57.0, 58.0]]]
        ),
        name=c.REACH,
    ),
    frequency=xr.DataArray(
        coords={
            c.GEO: _GEO_IDS,
            c.MEDIA_TIME: _TIME_STRS,
            c.RF_CHANNEL: _RF_CHANNEL_PAID,
        },
        data=np.array([[[1.1, 1.2], [2, 3]], [[4, 5], [6, 7]]]),
        name=c.FREQUENCY,
    ),
    rf_spend=xr.DataArray(
        coords={
            c.GEO: _GEO_IDS,
            c.TIME: _TIME_STRS,
            c.RF_CHANNEL: _RF_CHANNEL_PAID,
        },
        data=np.array([[[252, 253], [254, 255]], [[256, 257], [258, 259]]]),
        name=c.RF_SPEND,
    ),
    rf_spend_has_geo_dimension=True,
    rf_spend_has_time_dimension=True,
    kpi=xr.DataArray(
        coords={c.GEO: _GEO_IDS, c.TIME: _TIME_STRS},
        data=np.array([[1, 2], [3, 4]]),
        name=c.KPI,
    ),
    revenue_per_kpi=xr.DataArray(
        coords={c.GEO: _GEO_IDS, c.TIME: _TIME_STRS},
        data=np.ones((2, 2)),
        name=c.REVENUE_PER_KPI,
    ),
    controls=xr.DataArray(
        coords={
            c.GEO: _GEO_IDS,
            c.TIME: _TIME_STRS,
            c.CONTROL_VARIABLE: _CONTROL_VARIABLES,
        },
        data=np.array([[[31, 32], [33, 34]], [[35, 36], [37, 38]]]),
        name=c.CONTROLS,
    ),
    organic_media=None,
    organic_reach=None,
    organic_frequency=None,
    non_media_treatments=None,
)

MOCK_PROTO_RF_PAID_GRANULAR_NOT_LAGGED = text_format.Parse(
    """
   marketing_data_points {
    geo_info {
      geo_id: "geo_0"
      population: 11
    }
    date_interval {
      start_date {
        year: 2021
        month: 2
        day: 1
      }
      end_date {
        year: 2021
        month: 2
        day: 8
      }
    }
    control_variables {
      name: "control_0"
      value: 31.0
    }
    control_variables {
      name: "control_1"
      value: 32.0
    }
    reach_frequency_variables {
      channel_name: "rf_ch_paid_0"
      reach: 51
      average_frequency: 1.1
      spend: 252.0
    }
    reach_frequency_variables {
      channel_name: "rf_ch_paid_1"
      reach: 52
      average_frequency: 1.2
      spend: 253.0
    }
    kpi {
      name: "revenue"
      revenue {
        value: 1.0
      }
    }
  }
  marketing_data_points {
    geo_info {
      geo_id: "geo_0"
      population: 11
    }
    date_interval {
      start_date {
        year: 2021
        month: 2
        day: 8
      }
      end_date {
        year: 2021
        month: 2
        day: 15
      }
    }
    control_variables {
      name: "control_0"
      value: 33.0
    }
    control_variables {
      name: "control_1"
      value: 34.0
    }
    reach_frequency_variables {
      channel_name: "rf_ch_paid_0"
      reach: 53
      average_frequency: 2.0
      spend: 254.0
    }
    reach_frequency_variables {
      channel_name: "rf_ch_paid_1"
      reach: 54
      average_frequency: 3.0
      spend: 255.0
    }
    kpi {
      name: "revenue"
      revenue {
        value: 2.0
      }
    }
  }
  marketing_data_points {
    geo_info {
      geo_id: "geo_1"
      population: 12
    }
    date_interval {
      start_date {
        year: 2021
        month: 2
        day: 1
      }
      end_date {
        year: 2021
        month: 2
        day: 8
      }
    }
    control_variables {
      name: "control_0"
      value: 35.0
    }
    control_variables {
      name: "control_1"
      value: 36.0
    }
    reach_frequency_variables {
      channel_name: "rf_ch_paid_0"
      reach: 55
      average_frequency: 4.0
      spend: 256.0
    }
    reach_frequency_variables {
      channel_name: "rf_ch_paid_1"
      reach: 56
      average_frequency: 5.0
      spend: 257.0
    }
    kpi {
      name: "revenue"
      revenue {
        value: 3.0
      }
    }
  }
  marketing_data_points {
    geo_info {
      geo_id: "geo_1"
      population: 12
    }
    date_interval {
      start_date {
        year: 2021
        month: 2
        day: 8
      }
      end_date {
        year: 2021
        month: 2
        day: 15
      }
    }
    control_variables {
      name: "control_0"
      value: 37.0
    }
    control_variables {
      name: "control_1"
      value: 38.0
    }
    reach_frequency_variables {
      channel_name: "rf_ch_paid_0"
      reach: 57
      average_frequency: 6.0
      spend: 258.0
    }
    reach_frequency_variables {
      channel_name: "rf_ch_paid_1"
      reach: 58
      average_frequency: 7.0
      spend: 259.0
    }
    kpi {
      name: "revenue"
      revenue {
        value: 4.0
      }
    }
  }
  metadata {
    time_dimensions {
      name: "time"
      dates {
        year: 2021
        month: 2
        day: 1
      }
      dates {
        year: 2021
        month: 2
        day: 8
      }
    }
    time_dimensions {
      name: "media_time"
      dates {
        year: 2021
        month: 2
        day: 1
      }
      dates {
        year: 2021
        month: 2
        day: 8
      }
    }
    channel_dimensions {
      name: "reach"
      channels: "rf_ch_paid_0"
      channels: "rf_ch_paid_1"
    }
    channel_dimensions {
      name: "frequency"
      channels: "rf_ch_paid_0"
      channels: "rf_ch_paid_1"
    }
    control_names: "control_0"
    control_names: "control_1"
    kpi_type: "revenue"
  }
    """,
    marketing_pb.MarketingData(),
)

# Reach and Frequency, Organic, Expanded, Lagged
MOCK_INPUT_DATA_RF_ORGANIC_EXPANDED_LAGGED = mock.MagicMock(
    kpi_type=c.REVENUE,
    geo=xr.DataArray(np.array(_GEO_IDS)),
    time=xr.DataArray(np.array(_TIME_STRS)),
    media_time=xr.DataArray(np.array(_MEDIA_TIME_STRS)),
    population=xr.DataArray(
        coords={c.GEO: _GEO_IDS}, data=np.array([11.1, 12.2]), name=c.POPULATION
    ),
    media=None,
    media_spend=None,
    reach=xr.DataArray(
        coords={
            c.GEO: _GEO_IDS,
            c.MEDIA_TIME: _MEDIA_TIME_STRS,
            c.RF_CHANNEL: _RF_CHANNEL_PAID,
        },
        data=np.array([[[1, 2], [3, 4], [5, 6]], [[7, 8], [9, 10], [11, 12]]]),
        name=c.REACH,
    ),
    frequency=xr.DataArray(
        coords={
            c.GEO: _GEO_IDS,
            c.MEDIA_TIME: _MEDIA_TIME_STRS,
            c.RF_CHANNEL: _RF_CHANNEL_PAID,
        },
        data=np.array([
            [[2.1, 2.2], [2.3, 2.4], [2.5, 2.6]],
            [[2.7, 2.8], [2.9, 3.0], [3.1, 3.2]],
        ]),
        name=c.FREQUENCY,
    ),
    rf_spend=xr.DataArray(
        coords={c.RF_CHANNEL: _RF_CHANNEL_PAID},
        data=np.array([1004, 1008]),
        name=c.RF_SPEND,
    ),
    rf_spend_has_geo_dimension=False,
    rf_spend_has_time_dimension=False,
    organic_media=None,
    organic_reach=xr.DataArray(
        coords={
            c.GEO: _GEO_IDS,
            c.MEDIA_TIME: _MEDIA_TIME_STRS,
            c.ORGANIC_RF_CHANNEL: _RF_CHANNEL_ORGANIC,
        },
        data=np.array(
            [[[51, 52], [53, 54], [55, 56]], [[57, 58], [59, 60], [61, 62]]]
        ),
        name=c.ORGANIC_REACH,
    ),
    organic_frequency=xr.DataArray(
        coords={
            c.GEO: _GEO_IDS,
            c.MEDIA_TIME: _MEDIA_TIME_STRS,
            c.ORGANIC_RF_CHANNEL: _RF_CHANNEL_ORGANIC,
        },
        data=np.array([
            [[1.1, 1.2], [1.3, 1.4], [1.5, 1.6]],
            [[1.7, 1.8], [1.9, 2.0], [2.1, 2.2]],
        ]),
        name=c.ORGANIC_FREQUENCY,
    ),
    non_media_treatments=None,
    kpi=xr.DataArray(
        coords={c.GEO: _GEO_IDS, c.TIME: _TIME_STRS},
        data=np.array([[2, 2], [3, 3]]),
        name=c.KPI,
    ),
    revenue_per_kpi=xr.DataArray(
        coords={c.GEO: _GEO_IDS, c.TIME: _TIME_STRS},
        data=np.ones((2, 2)),
        name=c.REVENUE_PER_KPI,
    ),
    controls=xr.DataArray(
        coords={
            c.GEO: _GEO_IDS,
            c.TIME: _TIME_STRS,
            c.CONTROL_VARIABLE: _CONTROL_VARIABLES,
        },
        data=np.array([[[31, 32], [33, 34]], [[35, 36], [37, 38]]]),
        name=c.CONTROLS,
    ),
)

MOCK_PROTO_RF_ORGANIC_EXPANDED_LAGGED = text_format.Parse(
    """
    marketing_data_points {
      geo_info {
        geo_id: "geo_0"
        population: 11
      }
      date_interval {
        start_date {
          year: 2021
          month: 1
          day: 25
        }
        end_date {
          year: 2021
          month: 2
          day: 1
        }
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_0"
        reach: 1
        average_frequency: 2.1
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_1"
        reach: 2
        average_frequency: 2.2
      }
      reach_frequency_variables {
        channel_name: "rf_ch_organic_0"
        reach: 51
        average_frequency: 1.1
      }
      reach_frequency_variables {
        channel_name: "rf_ch_organic_1"
        reach: 52
        average_frequency: 1.2
      }
    }
    marketing_data_points {
      geo_info {
        geo_id: "geo_0"
        population: 11
      }
      date_interval {
        start_date {
          year: 2021
          month: 2
          day: 1
        }
        end_date {
          year: 2021
          month: 2
          day: 8
        }
      }
      control_variables {
        name: "control_0"
        value: 31.0
      }
      control_variables {
        name: "control_1"
        value: 32.0
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_0"
        reach: 3
        average_frequency: 2.3
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_1"
        reach: 4
        average_frequency: 2.4
      }
      reach_frequency_variables {
        channel_name: "rf_ch_organic_0"
        reach: 53
        average_frequency: 1.3
      }
      reach_frequency_variables {
        channel_name: "rf_ch_organic_1"
        reach: 54
        average_frequency: 1.4
      }
      kpi {
        name: "revenue"
        revenue {
          value: 2.0
        }
      }
    }
    marketing_data_points {
      geo_info {
        geo_id: "geo_0"
        population: 11
      }
      date_interval {
        start_date {
          year: 2021
          month: 2
          day: 8
        }
        end_date {
          year: 2021
          month: 2
          day: 15
        }
      }
      control_variables {
        name: "control_0"
        value: 33.0
      }
      control_variables {
        name: "control_1"
        value: 34.0
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_0"
        reach: 5
        average_frequency: 2.5
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_1"
        reach: 6
        average_frequency: 2.6
      }
      reach_frequency_variables {
        channel_name: "rf_ch_organic_0"
        reach: 55
        average_frequency: 1.5
      }
      reach_frequency_variables {
        channel_name: "rf_ch_organic_1"
        reach: 56
        average_frequency: 1.6
      }
      kpi {
        name: "revenue"
        revenue {
          value: 2.0
        }
      }
    }
    marketing_data_points {
      geo_info {
        geo_id: "geo_1"
        population: 12
      }
      date_interval {
        start_date {
          year: 2021
          month: 1
          day: 25
        }
        end_date {
          year: 2021
          month: 2
          day: 1
        }
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_0"
        reach: 7
        average_frequency: 2.7
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_1"
        reach: 8
        average_frequency: 2.8
      }
      reach_frequency_variables {
        channel_name: "rf_ch_organic_0"
        reach: 57
        average_frequency: 1.7
      }
      reach_frequency_variables {
        channel_name: "rf_ch_organic_1"
        reach: 58
        average_frequency: 1.8
      }
    }
    marketing_data_points {
      geo_info {
        geo_id: "geo_1"
        population: 12
      }
      date_interval {
        start_date {
          year: 2021
          month: 2
          day: 1
        }
        end_date {
          year: 2021
          month: 2
          day: 8
        }
      }
      control_variables {
        name: "control_0"
        value: 35.0
      }
      control_variables {
        name: "control_1"
        value: 36.0
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_0"
        reach: 9
        average_frequency: 2.9
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_1"
        reach: 10
        average_frequency: 3.0
      }
      reach_frequency_variables {
        channel_name: "rf_ch_organic_0"
        reach: 59
        average_frequency: 1.9
      }
      reach_frequency_variables {
        channel_name: "rf_ch_organic_1"
        reach: 60
        average_frequency: 2.0
      }
      kpi {
        name: "revenue"
        revenue {
          value: 3.0
        }
      }
    }
    marketing_data_points {
      geo_info {
        geo_id: "geo_1"
        population: 12
      }
      date_interval {
        start_date {
          year: 2021
          month: 2
          day: 8
        }
        end_date {
          year: 2021
          month: 2
          day: 15
        }
      }
      control_variables {
        name: "control_0"
        value: 37.0
      }
      control_variables {
        name: "control_1"
        value: 38.0
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_0"
        reach: 11
        average_frequency: 3.1
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_1"
        reach: 12
        average_frequency: 3.2
      }
      reach_frequency_variables {
        channel_name: "rf_ch_organic_0"
        reach: 61
        average_frequency: 2.1
      }
      reach_frequency_variables {
        channel_name: "rf_ch_organic_1"
        reach: 62
        average_frequency: 2.2
      }
      kpi {
        name: "revenue"
        revenue {
          value: 3.0
        }
      }
    }
    marketing_data_points {
      date_interval {
        start_date {
          year: 2021
          month: 1
          day: 25
        }
        end_date {
          year: 2021
          month: 2
          day: 15
        }
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_0"
        spend: 1004.0
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_1"
        spend: 1008.0
      }
    }
    metadata {
      time_dimensions {
        name: "time"
        dates {
          year: 2021
          month: 2
          day: 1
        }
        dates {
          year: 2021
          month: 2
          day: 8
        }
      }
      time_dimensions {
        name: "media_time"
        dates {
          year: 2021
          month: 1
          day: 25
        }
        dates {
          year: 2021
          month: 2
          day: 1
        }
        dates {
          year: 2021
          month: 2
          day: 8
        }
      }
      channel_dimensions {
        name: "reach"
        channels: "rf_ch_paid_0"
        channels: "rf_ch_paid_1"
      }
      channel_dimensions {
        name: "frequency"
        channels: "rf_ch_paid_0"
        channels: "rf_ch_paid_1"
      }
      channel_dimensions {
        name: "organic_reach"
        channels: "rf_ch_organic_0"
        channels: "rf_ch_organic_1"
      }
      channel_dimensions {
        name: "organic_frequency"
        channels: "rf_ch_organic_0"
        channels: "rf_ch_organic_1"
      }
      control_names: "control_0"
      control_names: "control_1"
      kpi_type: "revenue"
    }
    """,
    marketing_pb.MarketingData(),
)

# Reach and Frequency, Organic, Granular, Not Lagged
MOCK_INPUT_DATA_RF_ORGANIC_GRANULAR_NOT_LAGGED = mock.MagicMock(
    kpi_type=c.REVENUE,
    geo=xr.DataArray(np.array(_GEO_IDS)),
    time=xr.DataArray(np.array(_TIME_STRS)),
    media_time=xr.DataArray(np.array(_TIME_STRS)),
    population=xr.DataArray(
        coords={c.GEO: _GEO_IDS}, data=np.array([11.1, 12.2]), name=c.POPULATION
    ),
    media=None,
    media_spend=None,
    reach=xr.DataArray(
        coords={
            c.GEO: _GEO_IDS,
            c.MEDIA_TIME: _TIME_STRS,
            c.RF_CHANNEL: _RF_CHANNEL_PAID,
        },
        data=np.array([[[1.0, 2.0], [3.0, 4.0]], [[5.0, 6.0], [7.0, 8.0]]]),
        name=c.REACH,
    ),
    frequency=xr.DataArray(
        coords={
            c.GEO: _GEO_IDS,
            c.MEDIA_TIME: _TIME_STRS,
            c.RF_CHANNEL: _RF_CHANNEL_PAID,
        },
        data=np.array([[[2.1, 2.2], [3, 4]], [[5, 6], [7, 8]]]),
        name=c.FREQUENCY,
    ),
    rf_spend=xr.DataArray(
        coords={
            c.GEO: _GEO_IDS,
            c.TIME: _TIME_STRS,
            c.RF_CHANNEL: _RF_CHANNEL_PAID,
        },
        data=np.array([[[252, 253], [254, 255]], [[256, 257], [258, 259]]]),
        name=c.RF_SPEND,
    ),
    rf_spend_has_geo_dimension=True,
    rf_spend_has_time_dimension=True,
    organic_media=None,
    organic_reach=xr.DataArray(
        coords={
            c.GEO: _GEO_IDS,
            c.MEDIA_TIME: _TIME_STRS,
            c.ORGANIC_RF_CHANNEL: _RF_CHANNEL_ORGANIC,
        },
        data=np.array(
            [[[51.0, 52.0], [53.0, 54.0]], [[55.0, 56.0], [57.0, 58.0]]]
        ),
        name=c.ORGANIC_REACH,
    ),
    organic_frequency=xr.DataArray(
        coords={
            c.GEO: _GEO_IDS,
            c.MEDIA_TIME: _TIME_STRS,
            c.ORGANIC_RF_CHANNEL: _RF_CHANNEL_ORGANIC,
        },
        data=np.array(
            [[[51.0, 52.0], [53.0, 54.0]], [[55.0, 56.0], [57.0, 58.0]]]
        ),
        name=c.ORGANIC_FREQUENCY,
    ),
    non_media_treatments=None,
    kpi=xr.DataArray(
        coords={c.GEO: _GEO_IDS, c.TIME: _TIME_STRS},
        data=np.array([[2, 2], [3, 3]]),
        name=c.KPI,
    ),
    revenue_per_kpi=xr.DataArray(
        coords={c.GEO: _GEO_IDS, c.TIME: _TIME_STRS},
        data=np.ones((2, 2)),
        name=c.REVENUE_PER_KPI,
    ),
    controls=xr.DataArray(
        coords={
            c.GEO: _GEO_IDS,
            c.TIME: _TIME_STRS,
            c.CONTROL_VARIABLE: _CONTROL_VARIABLES,
        },
        data=np.array([[[31, 32], [33, 34]], [[35, 36], [37, 38]]]),
        name=c.CONTROLS,
    ),
)

MOCK_PROTO_RF_ORGANIC_GRANULAR_NOT_LAGGED = text_format.Parse(
    """
    marketing_data_points {
      geo_info {
        geo_id: "geo_0"
        population: 11
      }
      date_interval {
        start_date {
          year: 2021
          month: 2
          day: 1
        }
        end_date {
          year: 2021
          month: 2
          day: 8
        }
      }
      control_variables {
        name: "control_0"
        value: 31.0
      }
      control_variables {
        name: "control_1"
        value: 32.0
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_0"
        reach: 1
        average_frequency: 2.1
        spend: 252.0
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_1"
        reach: 2
        average_frequency: 2.2
        spend: 253.0
      }
      reach_frequency_variables {
        channel_name: "rf_ch_organic_0"
        reach: 51
        average_frequency: 51.0
      }
      reach_frequency_variables {
        channel_name: "rf_ch_organic_1"
        reach: 52
        average_frequency: 52.0
      }
      kpi {
        name: "revenue"
        revenue {
          value: 2.0
        }
      }
    }
    marketing_data_points {
      geo_info {
        geo_id: "geo_0"
        population: 11
      }
      date_interval {
        start_date {
          year: 2021
          month: 2
          day: 8
        }
        end_date {
          year: 2021
          month: 2
          day: 15
        }
      }
      control_variables {
        name: "control_0"
        value: 33.0
      }
      control_variables {
        name: "control_1"
        value: 34.0
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_0"
        reach: 3
        average_frequency: 3.0
        spend: 254.0
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_1"
        reach: 4
        average_frequency: 4.0
        spend: 255.0
      }
      reach_frequency_variables {
        channel_name: "rf_ch_organic_0"
        reach: 53
        average_frequency: 53.0
      }
      reach_frequency_variables {
        channel_name: "rf_ch_organic_1"
        reach: 54
        average_frequency: 54.0
      }
      kpi {
        name: "revenue"
        revenue {
          value: 2.0
        }
      }
    }
    marketing_data_points {
      geo_info {
        geo_id: "geo_1"
        population: 12
      }
      date_interval {
        start_date {
          year: 2021
          month: 2
          day: 1
        }
        end_date {
          year: 2021
          month: 2
          day: 8
        }
      }
      control_variables {
        name: "control_0"
        value: 35.0
      }
      control_variables {
        name: "control_1"
        value: 36.0
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_0"
        reach: 5
        average_frequency: 5.0
        spend: 256.0
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_1"
        reach: 6
        average_frequency: 6.0
        spend: 257.0
      }
      reach_frequency_variables {
        channel_name: "rf_ch_organic_0"
        reach: 55
        average_frequency: 55.0
      }
      reach_frequency_variables {
        channel_name: "rf_ch_organic_1"
        reach: 56
        average_frequency: 56.0
      }
      kpi {
        name: "revenue"
        revenue {
          value: 3.0
        }
      }
    }
    marketing_data_points {
      geo_info {
        geo_id: "geo_1"
        population: 12
      }
      date_interval {
        start_date {
          year: 2021
          month: 2
          day: 8
        }
        end_date {
          year: 2021
          month: 2
          day: 15
        }
      }
      control_variables {
        name: "control_0"
        value: 37.0
      }
      control_variables {
        name: "control_1"
        value: 38.0
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_0"
        reach: 7
        average_frequency: 7.0
        spend: 258.0
      }
      reach_frequency_variables {
        channel_name: "rf_ch_paid_1"
        reach: 8
        average_frequency: 8.0
        spend: 259.0
      }
      reach_frequency_variables {
        channel_name: "rf_ch_organic_0"
        reach: 57
        average_frequency: 57.0
      }
      reach_frequency_variables {
        channel_name: "rf_ch_organic_1"
        reach: 58
        average_frequency: 58.0
      }
      kpi {
        name: "revenue"
        revenue {
          value: 3.0
        }
      }
    }
    metadata {
      time_dimensions {
        name: "time"
        dates {
          year: 2021
          month: 2
          day: 1
        }
        dates {
          year: 2021
          month: 2
          day: 8
        }
      }
      time_dimensions {
        name: "media_time"
        dates {
          year: 2021
          month: 2
          day: 1
        }
        dates {
          year: 2021
          month: 2
          day: 8
        }
      }
      channel_dimensions {
        name: "reach"
        channels: "rf_ch_paid_0"
        channels: "rf_ch_paid_1"
      }
      channel_dimensions {
        name: "frequency"
        channels: "rf_ch_paid_0"
        channels: "rf_ch_paid_1"
      }
      channel_dimensions {
        name: "organic_reach"
        channels: "rf_ch_organic_0"
        channels: "rf_ch_organic_1"
      }
      channel_dimensions {
        name: "organic_frequency"
        channels: "rf_ch_organic_0"
        channels: "rf_ch_organic_1"
      }
      control_names: "control_0"
      control_names: "control_1"
      kpi_type: "revenue"
    }
    """,
    marketing_pb.MarketingData(),
)

MOCK_INPUT_DATA_NON_MEDIA_TREATMENTS = mock.MagicMock(
    kpi_type=c.REVENUE,
    geo=xr.DataArray(np.array(_GEO_IDS)),
    time=xr.DataArray(np.array(_TIME_STRS)),
    media_time=xr.DataArray(np.array(_MEDIA_TIME_STRS)),
    population=xr.DataArray(
        coords={c.GEO: _GEO_IDS}, data=np.array([11.1, 12.2]), name=c.POPULATION
    ),
    kpi=xr.DataArray(
        coords={c.GEO: _GEO_IDS, c.TIME: _TIME_STRS},
        data=np.array([[1, 2], [3, 4]]),
        name=c.KPI,
    ),
    revenue_per_kpi=xr.DataArray(
        coords={c.GEO: _GEO_IDS, c.TIME: _TIME_STRS},
        data=np.ones((2, 2)),
        name=c.REVENUE_PER_KPI,
    ),
    controls=xr.DataArray(
        coords={
            c.GEO: _GEO_IDS,
            c.TIME: _TIME_STRS,
            c.CONTROL_VARIABLE: _CONTROL_VARIABLES,
        },
        data=np.array([[[31, 32], [33, 34]], [[35, 36], [37, 38]]]),
        name=c.CONTROLS,
    ),
    non_media_treatments=xr.DataArray(
        coords={
            c.GEO: _GEO_IDS,
            c.TIME: _TIME_STRS,
            c.NON_MEDIA_CHANNEL: _NON_MEDIA_TREATMENT_VARIABLES,
        },
        data=np.array([[[61, 62], [63, 64]], [[65, 66], [67, 68]]]),
        name=c.NON_MEDIA_TREATMENTS,
    ),
    media=xr.DataArray(
        coords={
            c.GEO: _GEO_IDS,
            c.MEDIA_TIME: _MEDIA_TIME_STRS,
            c.MEDIA_CHANNEL: _MEDIA_CHANNEL_PAID,
        },
        data=np.array(
            [[[39, 40], [41, 42], [43, 44]], [[45, 46], [47, 48], [49, 50]]]
        ),
        name=c.MEDIA,
    ),
    media_spend=xr.DataArray(
        coords={c.MEDIA_CHANNEL: _MEDIA_CHANNEL_PAID},
        data=np.array([492, 496]),
        name=c.MEDIA_SPEND,
    ),
    media_spend_has_geo_dimension=False,
    media_spend_has_time_dimension=False,
    reach=None,
    frequency=None,
    rf_spend=None,
    rf_spend_has_geo_dimension=False,
    rf_spend_has_time_dimension=False,
    organic_media=None,
    organic_reach=None,
    organic_frequency=None,
)

MOCK_PROTO_NON_MEDIA_TREATMENTS = text_format.Parse(
    """
    marketing_data_points {
      geo_info {
        geo_id: "geo_0"
        population: 11
      }
      date_interval {
        start_date {
          year: 2021
          month: 1
          day: 25
        }
        end_date {
          year: 2021
          month: 2
          day: 1
        }
      }
      media_variables {
        channel_name: "ch_paid_0"
        scalar_metric {
          name: "impressions"
          value: 39.0
        }
      }
      media_variables {
        channel_name: "ch_paid_1"
        scalar_metric {
          name: "impressions"
          value: 40.0
        }
      }
    }
    marketing_data_points {
      geo_info {
        geo_id: "geo_0"
        population: 11
      }
      date_interval {
        start_date {
          year: 2021
          month: 2
          day: 1
        }
        end_date {
          year: 2021
          month: 2
          day: 8
        }
      }
      control_variables {
        name: "control_0"
        value: 31.0
      }
      control_variables {
        name: "control_1"
        value: 32.0
      }
      media_variables {
        channel_name: "ch_paid_0"
        scalar_metric {
          name: "impressions"
          value: 41.0
        }
      }
      media_variables {
        channel_name: "ch_paid_1"
        scalar_metric {
          name: "impressions"
          value: 42.0
        }
      }
      non_media_treatment_variables {
        name: "non_media_treatment_0"
        value: 61.0
      }
      non_media_treatment_variables {
        name: "non_media_treatment_1"
        value: 62.0
      }
      kpi {
        name: "revenue"
        revenue {
          value: 1.0
        }
      }
    }
    marketing_data_points {
      geo_info {
        geo_id: "geo_0"
        population: 11
      }
      date_interval {
        start_date {
          year: 2021
          month: 2
          day: 8
        }
        end_date {
          year: 2021
          month: 2
          day: 15
        }
      }
      control_variables {
        name: "control_0"
        value: 33.0
      }
      control_variables {
        name: "control_1"
        value: 34.0
      }
      media_variables {
        channel_name: "ch_paid_0"
        scalar_metric {
          name: "impressions"
          value: 43.0
        }
      }
      media_variables {
        channel_name: "ch_paid_1"
        scalar_metric {
          name: "impressions"
          value: 44.0
        }
      }
      non_media_treatment_variables {
        name: "non_media_treatment_0"
        value: 63.0
      }
      non_media_treatment_variables {
        name: "non_media_treatment_1"
        value: 64.0
      }
      kpi {
        name: "revenue"
        revenue {
          value: 2.0
        }
      }
    }
    marketing_data_points {
      geo_info {
        geo_id: "geo_1"
        population: 12
      }
      date_interval {
        start_date {
          year: 2021
          month: 1
          day: 25
        }
        end_date {
          year: 2021
          month: 2
          day: 1
        }
      }
      media_variables {
        channel_name: "ch_paid_0"
        scalar_metric {
          name: "impressions"
          value: 45.0
        }
      }
      media_variables {
        channel_name: "ch_paid_1"
        scalar_metric {
          name: "impressions"
          value: 46.0
        }
      }
    }
    marketing_data_points {
      geo_info {
        geo_id: "geo_1"
        population: 12
      }
      date_interval {
        start_date {
          year: 2021
          month: 2
          day: 1
        }
        end_date {
          year: 2021
          month: 2
          day: 8
        }
      }
      control_variables {
        name: "control_0"
        value: 35.0
      }
      control_variables {
        name: "control_1"
        value: 36.0
      }
      media_variables {
        channel_name: "ch_paid_0"
        scalar_metric {
          name: "impressions"
          value: 47.0
        }
      }
      media_variables {
        channel_name: "ch_paid_1"
        scalar_metric {
          name: "impressions"
          value: 48.0
        }
      }
      non_media_treatment_variables {
        name: "non_media_treatment_0"
        value: 65.0
      }
      non_media_treatment_variables {
        name: "non_media_treatment_1"
        value: 66.0
      }
      kpi {
        name: "revenue"
        revenue {
          value: 3.0
        }
      }
    }
    marketing_data_points {
      geo_info {
        geo_id: "geo_1"
        population: 12
      }
      date_interval {
        start_date {
          year: 2021
          month: 2
          day: 8
        }
        end_date {
          year: 2021
          month: 2
          day: 15
        }
      }
      control_variables {
        name: "control_0"
        value: 37.0
      }
      control_variables {
        name: "control_1"
        value: 38.0
      }
      media_variables {
        channel_name: "ch_paid_0"
        scalar_metric {
          name: "impressions"
          value: 49.0
        }
      }
      media_variables {
        channel_name: "ch_paid_1"
        scalar_metric {
          name: "impressions"
          value: 50.0
        }
      }
      non_media_treatment_variables {
        name: "non_media_treatment_0"
        value: 67.0
      }
      non_media_treatment_variables {
        name: "non_media_treatment_1"
        value: 68.0
      }
      kpi {
        name: "revenue"
        revenue {
          value: 4.0
        }
      }
    }
    marketing_data_points {
      date_interval {
        start_date {
          year: 2021
          month: 1
          day: 25
        }
        end_date {
          year: 2021
          month: 2
          day: 15
        }
      }
      media_variables {
        channel_name: "ch_paid_0"
        media_spend: 492.0
      }
      media_variables {
        channel_name: "ch_paid_1"
        media_spend: 496.0
      }
    }
    metadata {
      time_dimensions {
        name: "time"
        dates {
          year: 2021
          month: 2
          day: 1
        }
        dates {
          year: 2021
          month: 2
          day: 8
        }
      }
      time_dimensions {
        name: "media_time"
        dates {
          year: 2021
          month: 1
          day: 25
        }
        dates {
          year: 2021
          month: 2
          day: 1
        }
        dates {
          year: 2021
          month: 2
          day: 8
        }
      }
      channel_dimensions {
        name: "media"
        channels: "ch_paid_0"
        channels: "ch_paid_1"
      }
      control_names: "control_0"
      control_names: "control_1"
      non_media_treatment_names: "non_media_treatment_0"
      non_media_treatment_names: "non_media_treatment_1"
      kpi_type: "revenue"
    }
    """,
    marketing_pb.MarketingData(),
)

MOCK_INPUT_DATA_NO_REVENUE_PER_KPI = mock.MagicMock(
    kpi_type=c.NON_REVENUE,
    geo=xr.DataArray(np.array(_GEO_IDS)),
    time=xr.DataArray(np.array(_TIME_STRS)),
    media_time=xr.DataArray(np.array(_TIME_STRS)),
    population=xr.DataArray(
        coords={c.GEO: _GEO_IDS},
        data=np.array([1000.0, 1200.0]),
        name=c.POPULATION,
    ),
    kpi=xr.DataArray(
        coords={c.GEO: _GEO_IDS, c.TIME: _TIME_STRS},
        data=np.array([[50, 60], [70, 80]]),
        name=c.KPI,
    ),
    revenue_per_kpi=None,
    controls=xr.DataArray(
        coords={
            c.GEO: _GEO_IDS,
            c.TIME: _TIME_STRS,
            c.CONTROL_VARIABLE: _CONTROL_VARIABLES,
        },
        data=np.array([[[31, 32], [33, 34]], [[35, 36], [37, 38]]]),
        name=c.CONTROLS,
    ),
    media=xr.DataArray(
        coords={
            c.GEO: _GEO_IDS,
            c.MEDIA_TIME: _TIME_STRS,
            c.MEDIA_CHANNEL: _MEDIA_CHANNEL_PAID,
        },
        data=np.array([[[39, 40], [41, 42]], [[43, 44], [45, 46]]]),
        name=c.MEDIA,
    ),
    media_spend=xr.DataArray(
        coords={c.MEDIA_CHANNEL: _MEDIA_CHANNEL_PAID},
        data=np.array([492, 496]),
        name=c.MEDIA_SPEND,
    ),
    media_spend_has_geo_dimension=False,
    media_spend_has_time_dimension=False,
    reach=None,
    frequency=None,
    rf_spend=None,
    organic_media=None,
    organic_reach=None,
    organic_frequency=None,
    non_media_treatments=None,
)

# Expected Protobuf (Textproto format)
MOCK_PROTO_NO_REVENUE_PER_KPI = text_format.Parse(
    """
    marketing_data_points {
      geo_info { geo_id: "geo_0" population: 1000 }
      date_interval {
        start_date { year: 2021 month: 2 day: 1 }
        end_date { year: 2021 month: 2 day: 8 }
      }
      control_variables { name: "control_0" value: 31.0 }
      control_variables { name: "control_1" value: 32.0 }
       media_variables {
        channel_name: "ch_paid_0"
        scalar_metric {
          name: "impressions"
          value: 39.0
        }
       }
        media_variables {
        channel_name: "ch_paid_1"
        scalar_metric {
          name: "impressions"
          value: 40.0
        }
       }
      kpi { name: "non_revenue" non_revenue { value: 50.0 } }
    }
    marketing_data_points {
      geo_info { geo_id: "geo_0" population: 1000 }
      date_interval {
        start_date { year: 2021 month: 2 day: 8 }
        end_date { year: 2021 month: 2 day: 15 }
      }
      control_variables { name: "control_0" value: 33.0 }
      control_variables { name: "control_1" value: 34.0 }
     media_variables {
        channel_name: "ch_paid_0"
        scalar_metric {
          name: "impressions"
          value: 41.0
        }
      }
      media_variables {
        channel_name: "ch_paid_1"
        scalar_metric {
          name: "impressions"
          value: 42.0
        }
      }
      kpi { name: "non_revenue" non_revenue { value: 60.0 } }
    }
    marketing_data_points {
      geo_info { geo_id: "geo_1" population: 1200 }
      date_interval {
        start_date { year: 2021 month: 2 day: 1 }
        end_date { year: 2021 month: 2 day: 8 }
      }
      control_variables { name: "control_0" value: 35.0 }
      control_variables { name: "control_1" value: 36.0 }
            media_variables {
        channel_name: "ch_paid_0"
        scalar_metric {
          name: "impressions"
          value: 43.0
        }
      }
      media_variables {
        channel_name: "ch_paid_1"
        scalar_metric {
          name: "impressions"
          value: 44.0
        }
      }
      kpi { name: "non_revenue" non_revenue { value: 70.0 } }
    }
    marketing_data_points {
      geo_info {
        geo_id: "geo_1"
        population: 1200
      }
      date_interval {
        start_date {
          year: 2021
          month: 2
          day: 8
        }
        end_date {
          year: 2021
          month: 2
          day: 15
        }
      }
      control_variables {
        name: "control_0"
        value: 37.0
      }
      control_variables {
        name: "control_1"
        value: 38.0
      }
          media_variables {
        channel_name: "ch_paid_0"
        scalar_metric {
          name: "impressions"
          value: 45.0
        }
      }
      media_variables {
        channel_name: "ch_paid_1"
        scalar_metric {
          name: "impressions"
          value: 46.0
        }
      }
      kpi { name: "non_revenue" non_revenue { value: 80.0 } }
    }
    marketing_data_points {
        date_interval {
        start_date {
          year: 2021
          month: 2
          day: 1
        }
        end_date {
          year: 2021
          month: 2
          day: 15
        }
      }
      media_variables {
        channel_name: "ch_paid_0"
        media_spend: 492.0
      }
      media_variables {
        channel_name: "ch_paid_1"
        media_spend: 496.0
      }
    }
    metadata {
      time_dimensions { name: "time" dates { year: 2021 month: 2 day: 1 } dates { year: 2021 month: 2 day: 8} }
      time_dimensions { name: "media_time" dates { year: 2021 month: 2 day: 1 } dates { year: 2021 month: 2 day: 8 } }
      channel_dimensions { name: "media" channels: "ch_paid_0" channels: "ch_paid_1" }
      control_names: "control_0"
      control_names: "control_1"
      kpi_type: "non_revenue"
    }
    """,
    marketing_pb.MarketingData(),
)


# Hyperparameters test data
def get_default_model_spec() -> spec.ModelSpec:
  return spec.ModelSpec()


DEFAULT_HYPERPARAMETERS_PROTO = meridian_pb.Hyperparameters(
    media_effects_dist=_MediaEffectsDist.LOG_NORMAL,
    hill_before_adstock=False,
    max_lag=8,
    unique_sigma_for_each_geo=False,
    media_prior_type=_PaidMediaPriorType.PAID_MEDIA_PRIOR_TYPE_UNSPECIFIED,
    rf_prior_type=_PaidMediaPriorType.PAID_MEDIA_PRIOR_TYPE_UNSPECIFIED,
    paid_media_prior_type=_PaidMediaPriorType.PAID_MEDIA_PRIOR_TYPE_UNSPECIFIED,
    organic_media_prior_type=_NonPaidTreatmentsPriorType.NON_PAID_TREATMENTS_PRIOR_TYPE_CONTRIBUTION,
    organic_rf_prior_type=_NonPaidTreatmentsPriorType.NON_PAID_TREATMENTS_PRIOR_TYPE_CONTRIBUTION,
    non_media_treatments_prior_type=_NonPaidTreatmentsPriorType.NON_PAID_TREATMENTS_PRIOR_TYPE_CONTRIBUTION,
    enable_aks=False,
    global_adstock_decay='geometric',
)


def get_custom_model_spec_1() -> spec.ModelSpec:
  return spec.ModelSpec(
      prior=prior_distribution.PriorDistribution(),
      media_effects_dist=c.MEDIA_EFFECTS_NORMAL,
      hill_before_adstock=True,
      max_lag=777,
      unique_sigma_for_each_geo=True,
      media_prior_type=c.TREATMENT_PRIOR_TYPE_MROI,
      rf_prior_type=c.TREATMENT_PRIOR_TYPE_MROI,
      knots=2,
      baseline_geo='baseline_geo',
      roi_calibration_period=None,
      rf_roi_calibration_period=None,
      holdout_id=None,
      control_population_scaling_id=None,
      adstock_decay_spec='binomial',
  )


CUSTOM_HYPERPARAMETERS_PROTO_1 = meridian_pb.Hyperparameters(
    media_effects_dist=_MediaEffectsDist.NORMAL,
    hill_before_adstock=True,
    max_lag=777,
    unique_sigma_for_each_geo=True,
    media_prior_type=_PaidMediaPriorType.MROI,
    rf_prior_type=_PaidMediaPriorType.MROI,
    paid_media_prior_type=_PaidMediaPriorType.PAID_MEDIA_PRIOR_TYPE_UNSPECIFIED,
    organic_media_prior_type=_NonPaidTreatmentsPriorType.NON_PAID_TREATMENTS_PRIOR_TYPE_CONTRIBUTION,
    organic_rf_prior_type=_NonPaidTreatmentsPriorType.NON_PAID_TREATMENTS_PRIOR_TYPE_CONTRIBUTION,
    non_media_treatments_prior_type=_NonPaidTreatmentsPriorType.NON_PAID_TREATMENTS_PRIOR_TYPE_CONTRIBUTION,
    knots=[2],
    baseline_geo_string='baseline_geo',
    enable_aks=False,
    global_adstock_decay='binomial',
)


def get_custom_model_spec_2() -> spec.ModelSpec:
  return spec.ModelSpec(
      prior=prior_distribution.PriorDistribution(),
      media_effects_dist='log_normal',
      hill_before_adstock=True,
      max_lag=777,
      unique_sigma_for_each_geo=True,
      media_prior_type=c.TREATMENT_PRIOR_TYPE_ROI,
      rf_prior_type=c.TREATMENT_PRIOR_TYPE_ROI,
      organic_media_prior_type=c.TREATMENT_PRIOR_TYPE_CONTRIBUTION,
      organic_rf_prior_type=c.TREATMENT_PRIOR_TYPE_COEFFICIENT,
      non_media_treatments_prior_type=c.TREATMENT_PRIOR_TYPE_COEFFICIENT,
      non_media_baseline_values=['min', 0.5, 'max'],
      knots=[1, 5, 8],
      baseline_geo=3,
      roi_calibration_period=np.full((2, 3), True),
      rf_roi_calibration_period=np.full((4, 5), False),
      holdout_id=np.full((6,), True),
      control_population_scaling_id=np.full((7, 8), False),
      non_media_population_scaling_id=np.full((9, 10), False),
      adstock_decay_spec={'ch_paid_0': 'binomial', 'rf_ch_paid_1': 'geometric'},
  )

CUSTOM_HYPERPARAMETERS_PROTO_2 = meridian_pb.Hyperparameters(
    media_effects_dist=_MediaEffectsDist.LOG_NORMAL,
    hill_before_adstock=True,
    max_lag=777,
    unique_sigma_for_each_geo=True,
    media_prior_type=_PaidMediaPriorType.ROI,
    rf_prior_type=_PaidMediaPriorType.ROI,
    paid_media_prior_type=_PaidMediaPriorType.PAID_MEDIA_PRIOR_TYPE_UNSPECIFIED,
    organic_media_prior_type=_NonPaidTreatmentsPriorType.NON_PAID_TREATMENTS_PRIOR_TYPE_CONTRIBUTION,
    organic_rf_prior_type=_NonPaidTreatmentsPriorType.NON_PAID_TREATMENTS_PRIOR_TYPE_COEFFICIENT,
    non_media_treatments_prior_type=_NonPaidTreatmentsPriorType.NON_PAID_TREATMENTS_PRIOR_TYPE_COEFFICIENT,
    knots=[1, 5, 8],
    baseline_geo_int=3,
    roi_calibration_period=make_tensor_proto(
        dims=[2, 3],
        dtype=types_pb2.DT_BOOL,
        bool_vals=[True] * (2 * 3),
    ),
    rf_roi_calibration_period=make_tensor_proto(
        dims=[4, 5],
        dtype=types_pb2.DT_BOOL,
        bool_vals=[False] * (4 * 5),
    ),
    holdout_id=make_tensor_proto(
        dims=[6],
        dtype=types_pb2.DT_BOOL,
        bool_vals=[True] * 6,
    ),
    control_population_scaling_id=make_tensor_proto(
        dims=[7, 8],
        dtype=types_pb2.DT_BOOL,
        bool_vals=[False] * (7 * 8),
    ),
    non_media_population_scaling_id=make_tensor_proto(
        dims=[9, 10],
        dtype=types_pb2.DT_BOOL,
        bool_vals=[False] * (9 * 10),
    ),
    non_media_baseline_values=[
        meridian_pb.NonMediaBaselineValue(
            function_value=_NonMediaBaselineFunction.MIN
        ),
        meridian_pb.NonMediaBaselineValue(value=0.5),
        meridian_pb.NonMediaBaselineValue(
            function_value=_NonMediaBaselineFunction.MAX
        ),
    ],
    enable_aks=False,
    adstock_decay_by_channel=meridian_pb.AdstockDecayByChannel(
        channel_decays={'ch_paid_0': 'binomial', 'rf_ch_paid_1': 'geometric'}
    ),
)


def get_custom_model_spec_3() -> spec.ModelSpec:
  return spec.ModelSpec(
      prior=prior_distribution.PriorDistribution(),
      media_effects_dist=c.MEDIA_EFFECTS_NORMAL,
      hill_before_adstock=True,
      max_lag=777,
      unique_sigma_for_each_geo=True,
      media_prior_type=c.TREATMENT_PRIOR_TYPE_MROI,
      rf_prior_type=c.TREATMENT_PRIOR_TYPE_MROI,
      organic_media_prior_type=c.TREATMENT_PRIOR_TYPE_CONTRIBUTION,
      organic_rf_prior_type=c.TREATMENT_PRIOR_TYPE_CONTRIBUTION,
      non_media_treatments_prior_type=c.TREATMENT_PRIOR_TYPE_CONTRIBUTION,
      baseline_geo='baseline_geo',
      roi_calibration_period=None,
      rf_roi_calibration_period=None,
      holdout_id=None,
      control_population_scaling_id=None,
      enable_aks=True,
  )


CUSTOM_HYPERPARAMETERS_PROTO_3 = meridian_pb.Hyperparameters(
    media_effects_dist=_MediaEffectsDist.NORMAL,
    hill_before_adstock=True,
    max_lag=777,
    unique_sigma_for_each_geo=True,
    media_prior_type=_PaidMediaPriorType.MROI,
    rf_prior_type=_PaidMediaPriorType.MROI,
    paid_media_prior_type=_PaidMediaPriorType.PAID_MEDIA_PRIOR_TYPE_UNSPECIFIED,
    organic_media_prior_type=_NonPaidTreatmentsPriorType.NON_PAID_TREATMENTS_PRIOR_TYPE_CONTRIBUTION,
    organic_rf_prior_type=_NonPaidTreatmentsPriorType.NON_PAID_TREATMENTS_PRIOR_TYPE_CONTRIBUTION,
    non_media_treatments_prior_type=_NonPaidTreatmentsPriorType.NON_PAID_TREATMENTS_PRIOR_TYPE_CONTRIBUTION,
    baseline_geo_string='baseline_geo',
    enable_aks=True,
    global_adstock_decay='geometric',
)


def _create_tfp_params_from_dict(
    param_dict: dict[str, Any],
    distribution: backend.tfd.Distribution | backend.bijectors.Bijector,
) -> dict[str, meridian_pb.TfpParameterValue]:
  param_dict.update({
      'validate_args': False,
  })
  return {
      key: _create_tfp_param(key, value, distribution)
      for key, value in param_dict.items()
  }


def create_distribution_proto(
    distribution_type: str, **kwargs
) -> meridian_pb.TfpDistribution:
  distribution = getattr(backend.tfd, distribution_type)
  return meridian_pb.TfpDistribution(
      distribution_type=distribution_type,
      parameters=_create_tfp_params_from_dict(kwargs, distribution),
  )


def create_bijector_proto(
    bijector_type: str, **kwargs
) -> meridian_pb.TfpBijector:
  bijector = getattr(backend.bijectors, bijector_type)
  return meridian_pb.TfpBijector(
      bijector_type=bijector_type,
      parameters=_create_tfp_params_from_dict(kwargs, bijector),
  )


def _create_tfp_param(param_name, param_value, distribution):
  """Creates a TfpParameterValue object based on the input value's type."""
  match param_value:
    case float():
      return meridian_pb.TfpParameterValue(scalar_value=param_value)
    case int():
      return meridian_pb.TfpParameterValue(int_value=param_value)
    case bool():
      return meridian_pb.TfpParameterValue(bool_value=param_value)
    case str():
      return meridian_pb.TfpParameterValue(string_value=param_value)
    case None:
      return meridian_pb.TfpParameterValue(none_value=True)
    case list():
      value_generator = (
          _create_tfp_param(param_name, v, distribution) for v in param_value
      )
      tfp_list_value = meridian_pb.TfpParameterValue.List(
          values=value_generator
      )
      return meridian_pb.TfpParameterValue(list_value=tfp_list_value)
    case dict():
      dict_value = {
          key: _create_tfp_param(key, v, distribution)
          for key, v in param_value.items()
      }
      return meridian_pb.TfpParameterValue(dict_value=dict_value)
    case _ if isinstance(param_value, (np.ndarray, backend.Tensor)):
      return meridian_pb.TfpParameterValue(
          tensor_value=backend.make_tensor_proto(param_value)
      )
    case meridian_pb.TfpDistribution():
      return meridian_pb.TfpParameterValue(distribution_value=param_value)
    case meridian_pb.TfpBijector():
      return meridian_pb.TfpParameterValue(bijector_value=param_value)
    case backend.tfd.ReparameterizationType():
      fully_reparameterized = param_value == backend.tfd.FULLY_REPARAMETERIZED
      return meridian_pb.TfpParameterValue(
          fully_reparameterized=fully_reparameterized
      )
    case types.FunctionType():
      # Add custom functions used for tests.
      test_registry = {'distribution_fn': distribution_fn}

      for function_key, func in test_registry.items():
        if func == param_value:  # pylint: disable=comparison-with-callable
          return meridian_pb.TfpParameterValue(
              function_param=meridian_pb.TfpParameterValue.FunctionParam(
                  function_key=function_key
              )
          )
      # Function has default value.
      signature = inspect.signature(distribution.__init__)
      param = signature.parameters[param_name]
      if param.default:
        return meridian_pb.TfpParameterValue(
            function_param=meridian_pb.TfpParameterValue.FunctionParam(
                uses_default=True
            )
        )
      raise TypeError(
          f'No function found in registry for "{param_value.__name__}"'
      )
    case _:
      # Handle unsupported types.
      raise TypeError(f'Unsupported type: {type(param_value)}')


# Arbitrary function used for testing `tfd.Autoregressive`.
# https://github.com/tensorflow/probability/blob/65f265c62bb1e2d15ef3e25104afb245a6d52429/tensorflow_probability/python/distributions/autoregressive_test.py#L89
def distribution_fn(sample0):
  num_frames = sample0.shape[-1]
  mask = backend.one_hot(0, num_frames)[:, backend.newaxis]
  probs = backend.roll(backend.one_hot(sample0, 3), shift=1, axis=-2)
  probs = probs * (1.0 - mask) + backend.to_tensor([0.5, 0.5, 0]) * mask
  return backend.tfd.Independent(
      backend.tfd.Categorical(probs=probs), reinterpreted_batch_ndims=1
  )


def get_default_kwargs_split_fn():
  """Returns the default `kwargs_split_fn` used for tfd Distributions."""
  # `dist` can be any Distribution that has kwargs_split_fn in its signature.
  dist = backend.tfd.TransformedDistribution
  signature = inspect.signature(dist.__init__)
  kwargs_split_fn_param = signature.parameters['kwargs_split_fn']
  return kwargs_split_fn_param.default



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Serialization and deserialization of `EDASpec` objects."""

from __future__ import annotations

import warnings

from meridian.model.eda import eda_spec
from mmm.v1.model.meridian.eda import eda_spec_pb2 as eda_spec_pb
from schema.serde import function_registry as function_registry_utils
from schema.serde import serde


FunctionRegistry = function_registry_utils.FunctionRegistry
_FUNCTION_REGISTRY_NAME = "function_registry"


class EDASpecSerde(serde.Serde[eda_spec_pb.EDASpec, eda_spec.EDASpec]):
  """Serializes and deserializes an `EDASpec` object into an `EDASpec` proto."""

  def __init__(self, function_registry: FunctionRegistry):
    """Initializes an `EDASpecSerde` instance.

    Args:
      function_registry: A lookup table containing custom functions used by
        `EDASpec` objects. It's recommended to explicitly define the custom
        functions instead of using lambdas, as lambda functions may not be
        serialized successfully.
    """
    self._function_registry = function_registry

  @property
  def function_registry(self) -> FunctionRegistry:
    return self._function_registry

  def serialize(self, obj: eda_spec.EDASpec) -> eda_spec_pb.EDASpec:
    """Serializes the given `EDASpec` object into an `EDASpec` proto."""
    proto = eda_spec_pb.EDASpec(
        aggregation_config=self._to_aggregation_config_proto(
            obj.aggregation_config
        ),
        vif_spec=self._to_vif_spec_proto(obj.vif_spec),
    )
    hashed_function_registry = self.function_registry.hashed_registry
    proto.function_registry.update(hashed_function_registry)
    return proto

  def deserialize(
      self,
      serialized: eda_spec_pb.EDASpec,
      serialized_version: str = "",
      force_deserialization: bool = False,
  ) -> eda_spec.EDASpec:
    """Deserializes the `EDASpec` proto.

    Args:
      serialized: A serialized `EDASpec` object.
      serialized_version: The version of the serialized Meridian model. This is
        used to handle changes in deserialization logic across different
        versions.
      force_deserialization: If True, bypasses the safety check that validates
        whether functions within `function_registry` have changed after
        serialization. Use with caution.

    Returns:
      A deserialized `EDASpec` object.
    """
    if force_deserialization:
      warnings.warn(
          "You're attempting to deserialize an EDASpec while ignoring changes"
          " to custom functions. This is a risky operation that can"
          " potentially lead to a deserialized EDASpec that behaves"
          " differently from the original, resulting in unexpected behavior."
          " Please proceed with caution."
      )
    else:
      hashed_function_registry = getattr(serialized, _FUNCTION_REGISTRY_NAME)
      try:
        self.function_registry.validate(hashed_function_registry)
      except Exception as e:
        raise ValueError(
            f"An issue found during deserializing EDASpec: {e}"
        ) from e

    aggregation_config = self._from_aggregation_config_proto(
        serialized.aggregation_config
    )
    vif_spec = self._from_vif_spec_proto(serialized.vif_spec)
    return eda_spec.EDASpec(
        aggregation_config=aggregation_config,
        vif_spec=vif_spec,
    )

  def _to_aggregation_config_proto(
      self, config: eda_spec.AggregationConfig
  ) -> eda_spec_pb.AggregationConfig:
    """Converts a Python `AggregationConfig` to a proto."""
    proto = eda_spec_pb.AggregationConfig()
    if config.control_variables:
      for key, func in config.control_variables.items():
        proto.control_variables[key].CopyFrom(
            self._to_aggregation_function_proto(func, key, "control_variables")
        )
    if config.non_media_treatments:
      for key, func in config.non_media_treatments.items():
        proto.non_media_treatments[key].CopyFrom(
            self._to_aggregation_function_proto(
                func, key, "non_media_treatments"
            )
        )
    return proto

  def _to_aggregation_function_proto(
      self, func: eda_spec.AggregationFn, key: str, field: str
  ) -> eda_spec_pb.AggregationFunction:
    """Converts a Python aggregation function to a proto."""
    function_key = self.function_registry.get_function_key(func)
    if function_key is not None:
      return eda_spec_pb.AggregationFunction(function_key=function_key)

    raise ValueError(
        f"Custom aggregation function `{key}` in `{field}` detected, but not"
        " found in registry. Please add custom functions to registry when"
        " saving models."
    )

  def _from_aggregation_config_proto(
      self, proto: eda_spec_pb.AggregationConfig
  ) -> eda_spec.AggregationConfig:
    """Converts a proto `AggregationConfig` to a Python object."""
    control_variables = {
        key: self._from_aggregation_function_proto(key, val)
        for key, val in proto.control_variables.items()
    }
    non_media_treatments = {
        key: self._from_aggregation_function_proto(key, val)
        for key, val in proto.non_media_treatments.items()
    }
    return eda_spec.AggregationConfig(
        control_variables=control_variables,
        non_media_treatments=non_media_treatments,
    )

  def _from_aggregation_function_proto(
      self, var_name: str, agg_func: eda_spec_pb.AggregationFunction
  ) -> eda_spec.AggregationFn:
    """Converts a proto `AggregationFunction` to a Python function."""
    if not agg_func.function_key:
      raise ValueError(
          "Function key is required in `AggregationFunction` proto message."
          f" The function key for {var_name} is empty."
      )

    if agg_func.function_key in self.function_registry:
      return self.function_registry[agg_func.function_key]

    raise ValueError(
        f"Function key `{agg_func.function_key}` not found in registry."
    )

  def _to_vif_spec_proto(
      self, vif_spec_obj: eda_spec.VIFSpec
  ) -> eda_spec_pb.VIFSpec:
    """Converts a Python `VIFSpec` to a proto."""
    return eda_spec_pb.VIFSpec(
        geo_threshold=vif_spec_obj.geo_threshold,
        overall_threshold=vif_spec_obj.overall_threshold,
        national_threshold=vif_spec_obj.national_threshold,
    )

  def _from_vif_spec_proto(
      self, vif_spec_proto: eda_spec_pb.VIFSpec
  ) -> eda_spec.VIFSpec:
    """Converts a proto `VIFSpec` to a Python object."""
    return eda_spec.VIFSpec(
        geo_threshold=vif_spec_proto.geo_threshold,
        overall_threshold=vif_spec_proto.overall_threshold,
        national_threshold=vif_spec_proto.national_threshold,
    )



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Serde for Hyperparameters."""

import warnings

from meridian import backend
from meridian import constants as c
from meridian.model import spec
from mmm.v1.model.meridian import meridian_model_pb2 as meridian_pb
from schema.serde import constants as sc
from schema.serde import serde
import numpy as np

_MediaEffectsDist = meridian_pb.MediaEffectsDistribution
_PaidMediaPriorType = meridian_pb.PaidMediaPriorType
_NonPaidTreatmentsPriorType = meridian_pb.NonPaidTreatmentsPriorType
_NonMediaBaselineFunction = (
    meridian_pb.NonMediaBaselineValue.NonMediaBaselineFunction
)


def _media_effects_dist_to_proto_enum(
    media_effect_dict: str,
) -> _MediaEffectsDist:
  match media_effect_dict:
    case c.MEDIA_EFFECTS_LOG_NORMAL:
      return _MediaEffectsDist.LOG_NORMAL
    case c.MEDIA_EFFECTS_NORMAL:
      return _MediaEffectsDist.NORMAL
    case _:
      return _MediaEffectsDist.MEDIA_EFFECTS_DISTRIBUTION_UNSPECIFIED


def _proto_enum_to_media_effects_dist(
    proto_enum: _MediaEffectsDist,
) -> str:
  """Converts a `_MediaEffectsDist` enum to its string representation."""
  match proto_enum:
    case _MediaEffectsDist.LOG_NORMAL:
      return c.MEDIA_EFFECTS_LOG_NORMAL
    case _MediaEffectsDist.NORMAL:
      return c.MEDIA_EFFECTS_NORMAL
    case _MediaEffectsDist.MEDIA_EFFECTS_DISTRIBUTION_UNSPECIFIED:
      warnings.warn(
          "Media effects distribution is unspecified. Resolving to"
          " 'log-normal'."
      )
      return c.MEDIA_EFFECTS_LOG_NORMAL
    case _:
      raise ValueError(
          "Unsupported MediaEffectsDistribution proto enum value:"
          f" {proto_enum}."
      )


def _paid_media_prior_type_to_proto_enum(
    paid_media_prior_type: str | None,
) -> _PaidMediaPriorType:
  """Converts a paid media prior type string to its proto enum."""
  if paid_media_prior_type is None:
    return _PaidMediaPriorType.PAID_MEDIA_PRIOR_TYPE_UNSPECIFIED
  try:
    return _PaidMediaPriorType.Value(paid_media_prior_type.upper())
  except ValueError:
    warnings.warn(
        f"Invalid paid media prior type: {paid_media_prior_type}. Resolving to"
        " PAID_MEDIA_PRIOR_TYPE_UNSPECIFIED."
    )
    return _PaidMediaPriorType.PAID_MEDIA_PRIOR_TYPE_UNSPECIFIED


def _proto_enum_to_paid_media_prior_type(
    proto_enum: _PaidMediaPriorType,
) -> str | None:
  """Converts a `_PaidMediaPriorType` enum to its string representation."""
  if proto_enum == _PaidMediaPriorType.PAID_MEDIA_PRIOR_TYPE_UNSPECIFIED:
    return None
  return _PaidMediaPriorType.Name(proto_enum).lower()


def _non_paid_prior_type_to_proto_enum(
    non_paid_prior_type: str,
) -> _NonPaidTreatmentsPriorType:
  """Converts a non-paid prior type string to its proto enum."""
  try:
    return _NonPaidTreatmentsPriorType.Value(
        f"NON_PAID_TREATMENTS_PRIOR_TYPE_{non_paid_prior_type.upper()}"
    )
  except ValueError:
    warnings.warn(
        f"Invalid non-paid prior type: {non_paid_prior_type}. Resolving to"
        " NON_PAID_TREATMENTS_PRIOR_TYPE_CONTRIBUTION."
    )
    return (
        _NonPaidTreatmentsPriorType.NON_PAID_TREATMENTS_PRIOR_TYPE_CONTRIBUTION
    )


def _proto_enum_to_non_paid_prior_type(
    proto_enum: _NonPaidTreatmentsPriorType,
) -> str:
  """Converts a `_NonPaidTreatmentsPriorType` enum to its string representation."""
  if (
      proto_enum
      == _NonPaidTreatmentsPriorType.NON_PAID_TREATMENTS_PRIOR_TYPE_UNSPECIFIED
  ):
    warnings.warn(
        "Non-paid prior type is unspecified. Resolving to 'contribution'."
    )
    return c.TREATMENT_PRIOR_TYPE_CONTRIBUTION
  return (
      _NonPaidTreatmentsPriorType.Name(proto_enum)
      .replace("NON_PAID_TREATMENTS_PRIOR_TYPE_", "")
      .lower()
  )


class HyperparametersSerde(
    serde.Serde[meridian_pb.Hyperparameters, spec.ModelSpec]
):
  """Serializes and deserializes a ModelSpec into a `Hyperparameters` proto.

  Note that this Serde only handles the Hyperparameters part of ModelSpec.
  The 'prior' attribute of ModelSpec is serialized/deserialized separately
  using DistributionSerde.
  """

  def serialize(self, obj: spec.ModelSpec) -> meridian_pb.Hyperparameters:
    """Serializes the given ModelSpec into a `Hyperparameters` proto."""
    hyperparameters_proto = meridian_pb.Hyperparameters(
        media_effects_dist=_media_effects_dist_to_proto_enum(
            obj.media_effects_dist
        ),
        hill_before_adstock=obj.hill_before_adstock,
        unique_sigma_for_each_geo=obj.unique_sigma_for_each_geo,
        media_prior_type=_paid_media_prior_type_to_proto_enum(
            obj.media_prior_type
        ),
        rf_prior_type=_paid_media_prior_type_to_proto_enum(obj.rf_prior_type),
        paid_media_prior_type=_paid_media_prior_type_to_proto_enum(
            obj.paid_media_prior_type
        ),
        organic_media_prior_type=_non_paid_prior_type_to_proto_enum(
            obj.organic_media_prior_type
        ),
        organic_rf_prior_type=_non_paid_prior_type_to_proto_enum(
            obj.organic_rf_prior_type
        ),
        non_media_treatments_prior_type=_non_paid_prior_type_to_proto_enum(
            obj.non_media_treatments_prior_type
        ),
        enable_aks=obj.enable_aks,
    )
    if obj.max_lag is not None:
      hyperparameters_proto.max_lag = obj.max_lag

    if isinstance(obj.knots, int):
      hyperparameters_proto.knots.append(obj.knots)
    elif isinstance(obj.knots, list):
      hyperparameters_proto.knots.extend(obj.knots)

    if isinstance(obj.baseline_geo, str):
      hyperparameters_proto.baseline_geo_string = obj.baseline_geo
    elif isinstance(obj.baseline_geo, int):
      hyperparameters_proto.baseline_geo_int = obj.baseline_geo

    if obj.roi_calibration_period is not None:
      hyperparameters_proto.roi_calibration_period.CopyFrom(
          backend.make_tensor_proto(np.array(obj.roi_calibration_period))
      )
    if obj.rf_roi_calibration_period is not None:
      hyperparameters_proto.rf_roi_calibration_period.CopyFrom(
          backend.make_tensor_proto(np.array(obj.rf_roi_calibration_period))
      )
    if obj.holdout_id is not None:
      hyperparameters_proto.holdout_id.CopyFrom(
          backend.make_tensor_proto(np.array(obj.holdout_id))
      )
    if obj.control_population_scaling_id is not None:
      hyperparameters_proto.control_population_scaling_id.CopyFrom(
          backend.make_tensor_proto(np.array(obj.control_population_scaling_id))
      )
    if obj.non_media_population_scaling_id is not None:
      hyperparameters_proto.non_media_population_scaling_id.CopyFrom(
          backend.make_tensor_proto(
              np.array(obj.non_media_population_scaling_id)
          )
      )

    if isinstance(obj.adstock_decay_spec, str):
      hyperparameters_proto.global_adstock_decay = obj.adstock_decay_spec
    elif isinstance(obj.adstock_decay_spec, dict):
      hyperparameters_proto.adstock_decay_by_channel.channel_decays.update(
          obj.adstock_decay_spec
      )

    if obj.non_media_baseline_values is not None:
      for value in obj.non_media_baseline_values:
        value_proto = hyperparameters_proto.non_media_baseline_values.add()
        if isinstance(value, str):
          if value.lower() == "min":
            value_proto.function_value = _NonMediaBaselineFunction.MIN
          elif value.lower() == "max":
            value_proto.function_value = _NonMediaBaselineFunction.MAX
        elif isinstance(value, (float, int)):
          value_proto.value = float(value)

    return hyperparameters_proto

  def deserialize(
      self,
      serialized: meridian_pb.Hyperparameters,
      serialized_version: str = "",
  ) -> spec.ModelSpec:
    """Deserializes the given `Hyperparameters` proto into a ModelSpec.

    Note that this only deserializes the Hyperparameters part of ModelSpec.
    The 'prior' attribute of ModelSpec is deserialized separately
    using DistributionSerde and should be combined in the MeridianSerde.

    Args:
      serialized: The serialized `Hyperparameters` proto.
      serialized_version: The version of the serialized model. This is used to
        handle changes in deserialization logic across different versions.

    Returns:
      A Meridian model spec container.
    """
    baseline_geo = None
    baseline_geo_field = serialized.WhichOneof(sc.BASELINE_GEO_ONEOF)
    if baseline_geo_field == sc.BASELINE_GEO_INT:
      baseline_geo = serialized.baseline_geo_int
    elif baseline_geo_field == sc.BASELINE_GEO_STRING:
      baseline_geo = serialized.baseline_geo_string

    knots = None
    if serialized.knots:
      if len(serialized.knots) == 1:
        knots = serialized.knots[0]
      else:
        knots = list(serialized.knots)

    max_lag = serialized.max_lag if serialized.HasField(c.MAX_LAG) else None

    roi_calibration_period = (
        backend.make_ndarray(serialized.roi_calibration_period)
        if serialized.HasField(c.ROI_CALIBRATION_PERIOD)
        else None
    )
    rf_roi_calibration_period = (
        backend.make_ndarray(serialized.rf_roi_calibration_period)
        if serialized.HasField(c.RF_ROI_CALIBRATION_PERIOD)
        else None
    )

    holdout_id = (
        backend.make_ndarray(serialized.holdout_id)
        if serialized.HasField(sc.HOLDOUT_ID)
        else None
    )

    control_population_scaling_id = (
        backend.make_ndarray(serialized.control_population_scaling_id)
        if serialized.HasField(sc.CONTROL_POPULATION_SCALING_ID)
        else None
    )

    non_media_population_scaling_id = (
        backend.make_ndarray(serialized.non_media_population_scaling_id)
        if serialized.HasField(sc.NON_MEDIA_POPULATION_SCALING_ID)
        else None
    )

    non_media_baseline_values = None
    if serialized.non_media_baseline_values:
      non_media_baseline_values = []
      for value_proto in serialized.non_media_baseline_values:
        field = value_proto.WhichOneof("non_media_baseline_value")
        if field == "value":
          non_media_baseline_values.append(value_proto.value)
        elif field == "function_value":
          if value_proto.function_value == _NonMediaBaselineFunction.MIN:
            non_media_baseline_values.append("min")
          elif value_proto.function_value == _NonMediaBaselineFunction.MAX:
            non_media_baseline_values.append("max")
          elif (
              value_proto.function_value
              == _NonMediaBaselineFunction.NON_MEDIA_BASELINE_FUNCTION_UNSPECIFIED
          ):
            warnings.warn(
                "Non-media baseline function value is unspecified. Resolving to"
                " 'min'."
            )
            non_media_baseline_values.append("min")
          else:
            raise ValueError(
                "Unsupported NonMediaBaselineFunction proto enum value:"
                f" {value_proto.function_value}."
            )
        else:
          raise ValueError(
              f"Unsupported NonMediaBaselineValue proto enum value: {field}."
          )

    adstock_decay_spec_field = serialized.WhichOneof(sc.ADSTOCK_DECAY_SPEC)
    if adstock_decay_spec_field == sc.GLOBAL_ADSTOCK_DECAY:
      adstock_decay_spec = serialized.global_adstock_decay
    elif adstock_decay_spec_field == sc.ADSTOCK_DECAY_BY_CHANNEL:
      adstock_decay_spec = dict(
          serialized.adstock_decay_by_channel.channel_decays
      )
    else:
      adstock_decay_spec = sc.DEFAULT_DECAY

    return spec.ModelSpec(
        media_effects_dist=_proto_enum_to_media_effects_dist(
            serialized.media_effects_dist
        ),
        hill_before_adstock=serialized.hill_before_adstock,
        max_lag=max_lag,
        unique_sigma_for_each_geo=serialized.unique_sigma_for_each_geo,
        media_prior_type=_proto_enum_to_paid_media_prior_type(
            serialized.media_prior_type
        ),
        rf_prior_type=_proto_enum_to_paid_media_prior_type(
            serialized.rf_prior_type
        ),
        paid_media_prior_type=_proto_enum_to_paid_media_prior_type(
            serialized.paid_media_prior_type
        ),
        organic_media_prior_type=_proto_enum_to_non_paid_prior_type(
            serialized.organic_media_prior_type
        ),
        organic_rf_prior_type=_proto_enum_to_non_paid_prior_type(
            serialized.organic_rf_prior_type
        ),
        non_media_treatments_prior_type=_proto_enum_to_non_paid_prior_type(
            serialized.non_media_treatments_prior_type
        ),
        non_media_baseline_values=non_media_baseline_values,
        knots=knots,
        enable_aks=serialized.enable_aks,
        baseline_geo=baseline_geo,
        roi_calibration_period=roi_calibration_period,
        rf_roi_calibration_period=rf_roi_calibration_period,
        holdout_id=holdout_id,
        control_population_scaling_id=control_population_scaling_id,
        non_media_population_scaling_id=non_media_population_scaling_id,
        adstock_decay_spec=adstock_decay_spec,
    )



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Meridian EDA Outcome."""

import dataclasses
import enum
import typing
import pandas as pd
import xarray as xr

__all__ = [
    "EDASeverity",
    "EDAFinding",
    "AnalysisLevel",
    "AnalysisArtifact",
    "PairwiseCorrArtifact",
    "StandardDeviationArtifact",
    "VIFArtifact",
    "KpiInvariabilityArtifact",
    "CostPerMediaUnitArtifact",
    "EDACheckType",
    "ArtifactType",
    "EDAOutcome",
]


@enum.unique
class EDASeverity(enum.Enum):
  """Enumeration for the severity of an EDA check's finding."""

  # For the non-critical findings.
  INFO = enum.auto()
  # For the non-critical findings that require user attention.
  ATTENTION = enum.auto()
  # For unacceptable, model-blocking data errors.
  ERROR = enum.auto()


@dataclasses.dataclass(frozen=True)
class EDAFinding:
  """Encapsulates a single, specific finding from an EDA check.

  Attributes:
      severity: The severity level of the finding.
      explanation: A human-readable description about the EDA check and a
        potential actionable guidance on how to address or interpret this
        specific finding.
  """

  severity: EDASeverity
  explanation: str


@enum.unique
class AnalysisLevel(enum.Enum):
  """Enumeration for the level of an analysis.

  Attributes:
    OVERALL: Computed across all geos and time. When the analysis is performed
      on national data, this level is equivalent to the NATIONAL level.
    NATIONAL: Computed across time for data aggregated to the national level.
      When the analysis is performed on national data, this level is equivalent
      to the OVERALL level.
    GEO: Computed across time, for each geo.
  """

  OVERALL = enum.auto()
  NATIONAL = enum.auto()
  GEO = enum.auto()


@dataclasses.dataclass(frozen=True)
class AnalysisArtifact:
  """Base dataclass for analysis artifacts.

  Specific EDA artifacts should inherit from this class to store check-specific
  data for downstream processing (e.g., plotting).

  Attributes:
    level: The level of the analysis.
  """

  level: AnalysisLevel


@dataclasses.dataclass(frozen=True)
class PairwiseCorrArtifact(AnalysisArtifact):
  """Encapsulates artifacts from a single pairwise correlation analysis.

  Attributes:
    corr_matrix: Pairwise correlation matrix.
    extreme_corr_var_pairs: DataFrame of variable pairs exceeding the
      correlation threshold. Includes 'correlation' and 'abs_correlation'
      columns, and is sorted by 'abs_correlation' in descending order.
    extreme_corr_threshold: The threshold used to identify extreme correlation
      pairs.
  """

  corr_matrix: xr.DataArray
  extreme_corr_var_pairs: pd.DataFrame
  extreme_corr_threshold: float


@dataclasses.dataclass(frozen=True)
class StandardDeviationArtifact(AnalysisArtifact):
  """Encapsulates artifacts from a standard deviation analysis.

  Attributes:
    variable: The variable for which standard deviation is calculated.
    std_ds: Dataset with stdev_with_outliers and stdev_without_outliers.
    outlier_df: DataFrame with outliers.
  """

  variable: str
  std_ds: xr.Dataset
  outlier_df: pd.DataFrame


@dataclasses.dataclass(frozen=True)
class VIFArtifact(AnalysisArtifact):
  """Encapsulates artifacts from a single VIF analysis.

  Attributes:
    vif_da: DataArray with VIF values.
    outlier_df: DataFrame with extreme VIF values.
  """

  vif_da: xr.DataArray
  outlier_df: pd.DataFrame


@dataclasses.dataclass(frozen=True)
class KpiInvariabilityArtifact(AnalysisArtifact):
  """Encapsulates artifacts from a KPI invariability analysis.

  Attributes:
    kpi_da: DataArray of the KPI that is examined for variability.
    kpi_stdev: The standard deviation of the KPI, which is used to test the KPI
      invariability.
  """

  kpi_da: xr.DataArray
  kpi_stdev: xr.DataArray


@dataclasses.dataclass(frozen=True)
class CostPerMediaUnitArtifact(AnalysisArtifact):
  """Encapsulates artifacts from a Cost per Media Unit analysis.

  Attributes:
    cost_per_media_unit_da: DataArray of cost per media unit.
    cost_media_unit_inconsistency_df: DataFrame of time periods where cost and
      media units are inconsistent (e.g., zero cost with positive media units,
      or positive cost with zero media units).
    outlier_df: DataFrame with outliers of cost per media unit.
  """

  cost_per_media_unit_da: xr.DataArray
  cost_media_unit_inconsistency_df: pd.DataFrame
  outlier_df: pd.DataFrame


@enum.unique
class EDACheckType(enum.Enum):
  """Enumeration for the type of an EDA check."""

  PAIRWISE_CORRELATION = enum.auto()
  STANDARD_DEVIATION = enum.auto()
  MULTICOLLINEARITY = enum.auto()
  KPI_INVARIABILITY = enum.auto()
  COST_PER_MEDIA_UNIT = enum.auto()


ArtifactType = typing.TypeVar("ArtifactType", bound="AnalysisArtifact")


@dataclasses.dataclass(frozen=True)
class EDAOutcome(typing.Generic[ArtifactType]):
  """A dataclass for the outcomes of a single EDA check function.

  An EDA check function can discover multiple issues. This object groups all of
  those individual issues, reported as a list of `EDAFinding` objects.

  Attributes:
    check_type: The type of the EDA check that is being performed.
    findings: A list of all individual issues discovered by the check.
    analysis_artifacts: A list of analysis artifacts from the EDA check.
  """

  check_type: EDACheckType
  findings: list[EDAFinding]
  analysis_artifacts: list[ArtifactType]

  @property
  def get_geo_artifact(self) -> ArtifactType | None:
    """Returns the geo-level analysis artifact."""
    for artifact in self.analysis_artifacts:
      if artifact.level == AnalysisLevel.GEO:
        return artifact
    return None

  @property
  def get_national_artifact(self) -> ArtifactType | None:
    """Returns the national-level analysis artifact."""
    for artifact in self.analysis_artifacts:
      if artifact.level == AnalysisLevel.NATIONAL:
        return artifact
    return None
